"""
Script that computes the regulation trajectory of the rotor and the annual energy production

Nikhar J. Abbas, Pietro Bortolotti
January 2020
"""

import logging

import numpy as np
from openmdao.api import Group, ExplicitComponent
from scipy.optimize import brentq, minimize, minimize_scalar
from scipy.interpolate import PchipInterpolator

from wisdem.ccblade.Polar import Polar
from wisdem.ccblade.ccblade import CCBlade, CCAirfoil
from wisdem.commonse.utilities import smooth_abs, smooth_min, linspace_with_deriv
from wisdem.commonse.distribution import RayleighCDF, WeibullWithMeanCDF

logger = logging.getLogger("wisdem/weis")
TOL = 1e-3


class RotorPower(Group):
    def initialize(self):
        self.options.declare("modeling_options")
        self.options.declare("opt_options")

    def setup(self):
        modeling_options = self.options["modeling_options"]

        self.add_subsystem(
            "powercurve",
            RegulatedPowerCurve(modeling_options=modeling_options),
            promotes=[
                "v_min",
                "v_max",
                "rated_power",
                "omega_min",
                "omega_max",
                "control_maxTS",
                "tsr_operational",
                "control_pitch",
                "drivetrainType",
                "r",
                "chord",
                "theta",
                "Rhub",
                "Rtip",
                "hub_height",
                "precone",
                "tilt",
                "yaw",
                "precurve",
                "precurveTip",
                "presweep",
                "presweepTip",
                "airfoils_aoa",
                "airfoils_Re",
                "airfoils_cl",
                "airfoils_cd",
                "airfoils_cm",
                "nBlades",
                "rho",
                "mu",
                "shearExp",
                "hubloss",
                "tiploss",
                "wakerotation",
                "usecd",
                "nSector",
            ],
        )
        self.add_subsystem("gust", GustETM(std=modeling_options["WISDEM"]["RotorSE"]["gust_std"]))
        self.add_subsystem("cdf", WeibullWithMeanCDF(nspline=modeling_options["WISDEM"]["RotorSE"]["n_pc_spline"]))
        self.add_subsystem("aep", AEP(nspline=modeling_options["WISDEM"]["RotorSE"]["n_pc_spline"]), promotes=["AEP"])

        # Connections to the gust calculation
        self.connect("powercurve.rated_V", "gust.V_hub")

        # Connections to the Weibull CDF
        self.connect("powercurve.V_spline", "cdf.x")

        # Connections to the aep computation component
        self.connect("cdf.F", "aep.CDF_V")
        self.connect("powercurve.P_spline", "aep.P")


class GustETM(ExplicitComponent):
    # OpenMDAO component that generates an "equivalent gust" wind speed by summing an user-defined wind speed at hub height with 3 times sigma. sigma is the turbulent wind speed standard deviation for the extreme turbulence model, see IEC-61400-1 Eq. 19 paragraph 6.3.2.3

    def initialize(self):
        # number of standard deviations for strength of gust
        self.options.declare("std", default=3.0)

    def setup(self):
        # Inputs
        self.add_input("V_mean", val=0.0, units="m/s", desc="IEC average wind speed for turbine class")
        self.add_input("V_hub", val=0.0, units="m/s", desc="hub height wind speed")
        self.add_discrete_input("turbulence_class", val="A", desc="IEC turbulence class")

        # Output
        self.add_output("V_gust", val=0.0, units="m/s", desc="gust wind speed")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        V_mean = float(inputs["V_mean"][0])
        V_hub = float(inputs["V_hub"][0])
        std = self.options["std"]
        turbulence_class = discrete_inputs["turbulence_class"]

        if turbulence_class.upper() == "A":
            Iref = 0.16
        elif turbulence_class.upper() == "B":
            Iref = 0.14
        elif turbulence_class.upper() == "C":
            Iref = 0.12
        else:
            raise ValueError("Unknown Turbulence Class: " + str(turbulence_class) + " . Permitted values are A / B / C")

        c = 2.0
        sigma = c * Iref * (0.072 * (V_mean / c + 3) * (V_hub / c - 4) + 10)
        V_gust = V_hub + std * sigma
        outputs["V_gust"] = V_gust


class RegulatedPowerCurve(Group):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        modeling_options = self.options["modeling_options"]

        self.add_subsystem("compute_power_curve", ComputePowerCurve(modeling_options=modeling_options), promotes=["*"])

        self.add_subsystem("compute_splines", ComputeSplines(modeling_options=modeling_options), promotes=["*"])


class ComputePowerCurve(ExplicitComponent):
    """
    Iteratively call CCBlade to compute the power curve.
    """

    def initialize(self):
        self.options.declare("modeling_options")
        self.options.declare("debug", default=False)

    def setup(self):
        modeling_options = self.options["modeling_options"]
        self.n_span = n_span = modeling_options["WISDEM"]["RotorSE"]["n_span"]
        self.n_aoa = n_aoa = modeling_options["WISDEM"]["RotorSE"]["n_aoa"]  # Number of angle of attacks
        self.n_Re = n_Re = modeling_options["WISDEM"]["RotorSE"]["n_Re"]  # Number of Reynolds, so far hard set at 1
        self.n_tab = n_tab = modeling_options["WISDEM"]["RotorSE"][
            "n_tab"
        ]  # Number of tabulated data. For distributed aerodynamic control this could be > 1
        self.regulation_reg_III = modeling_options["WISDEM"]["RotorSE"]["regulation_reg_III"]
        self.n_pc = modeling_options["WISDEM"]["RotorSE"]["n_pc"]
        self.n_pc_spline = modeling_options["WISDEM"]["RotorSE"]["n_pc_spline"]

        # parameters
        self.add_input("v_min", val=0.0, units="m/s", desc="cut-in wind speed")
        self.add_input("v_max", val=0.0, units="m/s", desc="cut-out wind speed")
        self.add_input("rated_power", val=0.0, units="W", desc="electrical rated power")
        self.add_input("omega_min", val=0.0, units="rpm", desc="minimum allowed rotor rotation speed")
        self.add_input("omega_max", val=0.0, units="rpm", desc="maximum allowed rotor rotation speed")
        self.add_input("control_maxTS", val=0.0, units="m/s", desc="maximum allowed blade tip speed")
        self.add_input("tsr_operational", val=0.0, desc="tip-speed ratio in Region 2 (should be optimized externally)")
        self.add_input(
            "control_pitch",
            val=0.0,
            units="deg",
            desc="pitch angle in region 2 (and region 3 for fixed pitch machines)",
        )
        self.add_input("ps_percent", val=1.0, desc="Scalar applied to the max torque within RotorSE for peak thrust shaving. Only used if `peak_thrust_shaving` is True.")
        self.add_discrete_input("fix_pitch_regI12", val=False, desc="If True, pitch is fixed in region I1/2, i.e. when min rpm is enforced.")

        self.add_discrete_input("drivetrainType", val="GEARED")
        self.add_input("gearbox_efficiency", val=1.0)
        self.add_input(
            "generator_efficiency",
            val=np.ones(self.n_pc),
            desc="Generator efficiency at various rpm values to support table lookup",
        )
        self.add_input(
            "lss_rpm",
            val=np.zeros(self.n_pc),
            units="rpm",
            desc="Low speed shaft RPM values at which the generator efficiency values are given",
        )

        self.add_input(
            "r",
            val=np.zeros(n_span),
            units="m",
            desc="radial locations where blade is defined (should be increasing and not go all the way to hub or tip)",
        )
        self.add_input("chord", val=np.zeros(n_span), units="m", desc="chord length at each section")
        self.add_input(
            "theta",
            val=np.zeros(n_span),
            units="deg",
            desc="twist angle at each section (positive decreases angle of attack)",
        )
        self.add_input("Rhub", val=0.0, units="m", desc="hub radius")
        self.add_input("Rtip", val=0.0, units="m", desc="tip radius")
        self.add_input("hub_height", val=0.0, units="m", desc="hub height")
        self.add_input("precone", val=0.0, units="deg", desc="precone angle")
        self.add_input("tilt", val=0.0, units="deg", desc="shaft tilt")
        self.add_input("yaw", val=0.0, units="deg", desc="yaw error")
        self.add_input("precurve", val=np.zeros(n_span), units="m", desc="precurve at each section")
        self.add_input("precurveTip", val=0.0, units="m", desc="precurve at tip")
        self.add_input("presweep", val=np.zeros(n_span), units="m", desc="presweep at each section")
        self.add_input("presweepTip", val=0.0, units="m", desc="presweep at tip")

        # self.add_discrete_input('airfoils',  val=[0]*n_span,                      desc='CCAirfoil instances')
        self.add_input("airfoils_cl", val=np.zeros((n_span, n_aoa, n_Re, n_tab)), desc="lift coefficients, spanwise")
        self.add_input("airfoils_cd", val=np.zeros((n_span, n_aoa, n_Re, n_tab)), desc="drag coefficients, spanwise")
        self.add_input("airfoils_cm", val=np.zeros((n_span, n_aoa, n_Re, n_tab)), desc="moment coefficients, spanwise")
        self.add_input("airfoils_aoa", val=np.zeros((n_aoa)), units="deg", desc="angle of attack grid for polars")
        self.add_input("airfoils_Re", val=np.zeros((n_Re)), desc="Reynolds numbers of polars")
        self.add_discrete_input("nBlades", val=0, desc="number of blades")
        self.add_input("rho", val=1.225, units="kg/m**3", desc="density of air")
        self.add_input("mu", val=1.81e-5, units="kg/(m*s)", desc="dynamic viscosity of air")
        self.add_input("shearExp", val=0.0, desc="shear exponent")
        self.add_discrete_input(
            "nSector", val=4, desc="number of sectors to divide rotor face into in computing thrust and power"
        )
        self.add_discrete_input("tiploss", val=True, desc="include Prandtl tip loss model")
        self.add_discrete_input("hubloss", val=True, desc="include Prandtl hub loss model")
        self.add_discrete_input(
            "wakerotation",
            val=True,
            desc="include effect of wake rotation (i.e., tangential induction factor is nonzero)",
        )
        self.add_discrete_input("usecd", val=True, desc="use drag coefficient in computing induction factors")

        # outputs
        self.add_output("V", val=np.zeros(self.n_pc), units="m/s", desc="wind vector")
        self.add_output("Omega", val=np.zeros(self.n_pc), units="rpm", desc="rotor rotational speed")
        self.add_output("pitch", val=np.zeros(self.n_pc), units="deg", desc="rotor pitch schedule")
        self.add_output("P", val=np.zeros(self.n_pc), units="W", desc="rotor electrical power")
        self.add_output("P_aero", val=np.zeros(self.n_pc), units="W", desc="rotor mechanical power")
        self.add_output("T", val=np.zeros(self.n_pc), units="N", desc="rotor aerodynamic thrust")
        self.add_output("Q", val=np.zeros(self.n_pc), units="N*m", desc="rotor aerodynamic torque")
        self.add_output("M", val=np.zeros(self.n_pc), units="N*m", desc="blade root moment")
        self.add_output("Cp", val=np.zeros(self.n_pc), desc="rotor electrical power coefficient")
        self.add_output("Cp_aero", val=np.zeros(self.n_pc), desc="rotor aerodynamic power coefficient")
        self.add_output("Ct_aero", val=np.zeros(self.n_pc), desc="rotor aerodynamic thrust coefficient")
        self.add_output("Cq_aero", val=np.zeros(self.n_pc), desc="rotor aerodynamic torque coefficient")
        self.add_output("Cm_aero", val=np.zeros(self.n_pc), desc="rotor aerodynamic moment coefficient")
        self.add_output("ax_induct_rotor", val=np.zeros(self.n_pc), desc="rotor aerodynamic induction")

        self.add_output("V_R25", val=0.0, units="m/s", desc="region 2.5 transition wind speed")
        self.add_output("rated_V", val=0.0, units="m/s", desc="rated wind speed")
        self.add_output("rated_Omega", val=0.0, units="rpm", desc="rotor rotation speed at rated")
        self.add_output("rated_pitch", val=0.0, units="deg", desc="pitch setting at rated")
        self.add_output("rated_T", val=0.0, units="N", desc="rotor aerodynamic thrust at rated")
        self.add_output("rated_Q", val=0.0, units="N*m", desc="rotor aerodynamic torque at rated")
        self.add_output("rated_mech", val=0.0, units="W", desc="Mechanical shaft power at rated")
        self.add_output(
            "ax_induct_regII", val=np.zeros(n_span), desc="rotor axial induction at cut-in wind speed along blade span"
        )
        self.add_output(
            "tang_induct_regII",
            val=np.zeros(n_span),
            desc="rotor tangential induction at cut-in wind speed along blade span",
        )
        self.add_output(
            "aoa_regII",
            val=np.zeros(n_span),
            units="deg",
            desc="angle of attack distribution along blade span at cut-in wind speed",
        )
        self.add_output(
            "L_D",
            val=np.zeros(n_span),
            desc="Lift over drag distribution along blade span at cut-in wind speed",
        )
        self.add_output("Cp_regII", val=0.0, desc="power coefficient at cut-in wind speed")
        self.add_output("Ct_regII", val=0.0, desc="thrust coefficient at cut-in wind speed")
        self.add_output(
            "cl_regII", val=np.zeros(n_span), desc="lift coefficient distribution along blade span at cut-in wind speed"
        )
        self.add_output(
            "cd_regII", val=np.zeros(n_span), desc="drag coefficient distribution along blade span at cut-in wind speed"
        )
        self.add_output("rated_efficiency", val=1.0, desc="Efficiency at rated conditions")

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Saving out inputs for easy debugging of troublesome cases
        if self.options["debug"]:
            np.savez(
                "debug.npz",
                v_min=inputs["v_min"],
                v_max=inputs["v_max"],
                rated_power=inputs["rated_power"],
                omega_min=inputs["omega_min"],
                omega_max=inputs["omega_max"],
                control_maxTS=inputs["control_maxTS"],
                tsr_operational=inputs["tsr_operational"],
                control_pitch=inputs["control_pitch"],
                gearbox_efficiency=inputs["gearbox_efficiency"],
                generator_efficiency=inputs["generator_efficiency"],
                lss_rpm=inputs["lss_rpm"],
                r=inputs["r"],
                chord=inputs["chord"],
                theta=inputs["theta"],
                Rhub=inputs["Rhub"],
                Rtip=inputs["Rtip"],
                hub_height=inputs["hub_height"],
                precone=inputs["precone"],
                tilt=inputs["tilt"],
                yaw=inputs["yaw"],
                precurve=inputs["precurve"],
                precurveTip=inputs["precurveTip"],
                presweep=inputs["presweep"],
                presweepTip=inputs["presweepTip"],
                airfoils_cl=inputs["airfoils_cl"],
                airfoils_cd=inputs["airfoils_cd"],
                airfoils_cm=inputs["airfoils_cm"],
                airfoils_aoa=inputs["airfoils_aoa"],
                airfoils_Re=inputs["airfoils_Re"],
                rho=inputs["rho"],
                mu=inputs["mu"],
                shearExp=inputs["shearExp"],
                nBlades=discrete_inputs["nBlades"],
            )

        # Create Airfoil class instances
        af = [None] * self.n_span
        for i in range(self.n_span):
            if self.n_tab > 1:
                ref_tab = int(np.floor(self.n_tab / 2))
                af[i] = CCAirfoil(
                    inputs["airfoils_aoa"],
                    inputs["airfoils_Re"],
                    inputs["airfoils_cl"][i, :, :, ref_tab],
                    inputs["airfoils_cd"][i, :, :, ref_tab],
                    inputs["airfoils_cm"][i, :, :, ref_tab],
                )
            else:
                af[i] = CCAirfoil(
                    inputs["airfoils_aoa"],
                    inputs["airfoils_Re"],
                    inputs["airfoils_cl"][i, :, :, 0],
                    inputs["airfoils_cd"][i, :, :, 0],
                    inputs["airfoils_cm"][i, :, :, 0],
                )

        self.ccblade = CCBlade(
            inputs["r"],
            inputs["chord"],
            inputs["theta"],
            af,
            inputs["Rhub"][0],
            inputs["Rtip"][0],
            discrete_inputs["nBlades"],
            inputs["rho"][0],
            inputs["mu"][0],
            inputs["precone"][0],
            inputs["tilt"][0],
            inputs["yaw"][0],
            inputs["shearExp"][0],
            inputs["hub_height"][0],
            discrete_inputs["nSector"],
            inputs["precurve"],
            inputs["precurveTip"][0],
            inputs["presweep"],
            inputs["presweepTip"][0],
            discrete_inputs["tiploss"],
            discrete_inputs["hubloss"],
            discrete_inputs["wakerotation"],
            discrete_inputs["usecd"],
        )

        # JPJ: what is this grid for? Seems to be a special distribution of velocities
        # for the hub
        grid0 = np.cumsum(np.abs(np.diff(np.cos(np.linspace(-np.pi / 4.0, np.pi / 2.0, self.n_pc)))))
        grid1 = (grid0 - grid0[0]) / (grid0[-1] - grid0[0])
        Uhub = grid1 * (inputs["v_max"][0] - inputs["v_min"][0]) + inputs["v_min"][0]

        P_aero = np.zeros(Uhub.shape)
        Cp_aero = np.zeros(Uhub.shape)
        Ct_aero = np.zeros(Uhub.shape)
        Cq_aero = np.zeros(Uhub.shape)
        Cm_aero = np.zeros(Uhub.shape)
        P = np.zeros(Uhub.shape)
        Cp = np.zeros(Uhub.shape)
        T = np.zeros(Uhub.shape)
        Q = np.zeros(Uhub.shape)
        M = np.zeros(Uhub.shape)
        min_pitch = inputs["control_pitch"][0]
        pitch = np.zeros(Uhub.shape) + min_pitch

        # Unpack variables
        P_rated = float(inputs["rated_power"][0])
        # Rtip is passed without cone to CCBlade, which multiplies it by the cos(cone) internally, 
        # but it is also multiplied by cos(cone) to be used correctly in this script
        Rtip_cone = float(inputs["Rtip"][0]) * np.cos(np.deg2rad(float(inputs["precone"][0])))
        tsr = float(inputs["tsr_operational"][0])
        driveType = discrete_inputs["drivetrainType"]

        ## POWERCURVE PRELIMS ##
        # Set rotor speed based on TSR
        Omega_tsr = Uhub * tsr / Rtip_cone

        # Determine maximum rotor speed (rad/s)- either by TS or by control input
        Omega_max = min([float(inputs["control_maxTS"][0]) / Rtip_cone,
                         float(inputs["omega_max"][0]) * np.pi / 30.0])

        # Apply maximum and minimum rotor speed limits
        Omega_min = float(inputs["omega_min"][0]) * np.pi / 30.0
        Omega = np.maximum(np.minimum(Omega_tsr, Omega_max), Omega_min)
        Omega_rpm = Omega * 30.0 / np.pi

        # Create table lookup of total drivetrain efficiency, where rpm is first column and second column is gearbox*generator
        lss_rpm = inputs["lss_rpm"]
        gen_eff = inputs["generator_efficiency"]
        if not np.any(lss_rpm):
            lss_rpm = np.linspace(np.maximum(0.1, Omega_rpm[0]), Omega_rpm[-1], self.n_pc - 1)
            _, gen_eff = compute_P_and_eff(
                P_rated * lss_rpm / lss_rpm[-1],
                P_rated,
                np.zeros(self.n_pc - 1),
                driveType,
                np.zeros((self.n_pc - 1, 2)),
            )

        # driveEta  = np.c_[lss_rpm, float(inputs['gearbox_efficiency'])*gen_eff]
        driveEta = float(inputs["gearbox_efficiency"][0]) * gen_eff

        # Set baseline power production
        myout, derivs = self.ccblade.evaluate(Uhub, Omega_tsr * 30.0 / np.pi, pitch, coefficients=True)
        P_aero, T, Q, M, Cp_aero, Ct_aero, Cq_aero, Cm_aero = [
            myout[key] for key in ["P", "T", "Q", "Mb", "CP", "CT", "CQ", "CMb"]
        ]
        # P, eff  = compute_P_and_eff(P_aero, P_rated, Omega_rpm, driveType, driveEta)
        eff = np.interp(Omega_rpm, lss_rpm, driveEta)
        P = P_aero * eff
        Cp = Cp_aero * eff

        # Find rated index and guess at rated speed
        if P_aero[-1] > P_rated:
            U_rated = np.interp(P_rated, P, Uhub)
            i_rated = np.nonzero(U_rated <= Uhub)[0][0]
        else:
            U_rated = Uhub[-1] + 1e-6

        # Find Region 3 index
        found_rated = P_aero[-1] > P_rated
        region3 = len(np.nonzero(P >= P_rated)[0]) > 0

        # Guess at Region 2.5, but we will do a more rigorous search below
        if Omega_max < Omega_tsr[-1]:
            U_2p5 = np.interp(Omega[-1], Omega_tsr, Uhub)
            outputs["V_R25"] = U_2p5
        else:
            U_2p5 = U_rated
        region2p5 = U_2p5 < U_rated

        # Initialize peak shaving thrust value, will be updated later
        ps_percent = float(inputs["ps_percent"][0])
        if ps_percent < 1.:
            peak_thrust_shaving = True
        else:
            peak_thrust_shaving = False
        max_T = ps_percent * T.max() if peak_thrust_shaving and found_rated else 1e16

        ## REGION II.5 and RATED ##
        # Solve for rated velocity
        if found_rated:
            i = i_rated

            def const_Urated(x):
                pitch_i = x[0]
                Uhub_i = x[1]
                Omega_i = min([Uhub_i * tsr / Rtip_cone, Omega_max])
                Omega_i_rpm = Omega_i * 30.0 / np.pi
                myout, _ = self.ccblade.evaluate([Uhub_i], [Omega_i_rpm], [pitch_i], coefficients=False)
                P_aero_i = float(myout["P"][0])
                # P_i,_  = compute_P_and_eff(P_aero_i.flatten(), P_rated, Omega_i_rpm, driveType, driveEta)
                eff_i = np.interp(Omega_i_rpm, lss_rpm, driveEta)
                if isinstance(eff_i, type(np.array([]))): eff_i = eff_i[0]
                P_i = float(P_aero_i * eff_i)
                return 1e-4 * (P_i - P_rated)

            if region2p5:
                # Have to search over both pitch and speed
                x0 = [0.0, U_rated]
                imin = max(i - 3, 0)
                imax = min(i + 2, len(Uhub) - 1)
                bnds = [[min_pitch, 15.0], [Uhub[imin] + TOL, Uhub[imax] - TOL]]
                const = {}
                const["type"] = "eq"
                const["fun"] = const_Urated
                params_rated = minimize(
                    lambda x: x[1], x0, method="slsqp", bounds=bnds, constraints=const, tol=TOL, options={"maxiter": 20, "disp": False}
                )

                if params_rated.success and not np.isnan(params_rated.x[1]):
                    U_rated = params_rated.x[1]
                    pitch_rated = params_rated.x[0]
                else:
                    U_rated = U_rated  # Use guessed value earlier
                    pitch_rated = min_pitch
            else:
                # Just search over speed
                pitch_rated = min_pitch
                i_Uhub = np.max([0, i - 2])
                try:
                    U_rated = brentq(
                        lambda x: const_Urated([0.0, x]),
                        Uhub[i_Uhub],
                        Uhub[i + 2],
                        xtol=1e-1 * TOL,
                        rtol=1e-2 * TOL,
                        maxiter=40,
                        disp=False,
                    )
                except ValueError:
                    U_rated = minimize_scalar(
                        lambda x: np.abs(const_Urated([0.0, x])),
                        bounds=[Uhub[i_Uhub], Uhub[i + 2]],
                        method="bounded",
                        options={"disp": False, "xatol": TOL, "maxiter": 40},
                    )["x"]

            Omega_tsr_rated = U_rated * tsr / Rtip_cone
            Omega_rated = np.minimum(Omega_tsr_rated, Omega_max)
            Omega_rpm_rated = Omega_rated * 30.0 / np.pi
            myout, _ = self.ccblade.evaluate([U_rated], [Omega_rpm_rated], [pitch_rated], coefficients=True)
            (
                P_aero_rated,
                T_rated,
                Q_rated,
                M_rated,
                Cp_aero_rated,
                Ct_aero_rated,
                Cq_aero_rated,
                Cm_aero_rated,
            ) = [float(myout[key][0]) for key in ["P", "T", "Q", "Mb", "CP", "CT", "CQ", "CMb"]]
            eff_rated = np.interp(Omega_rpm_rated, lss_rpm, driveEta)
            Cp_rated = Cp_aero_rated * eff_rated
            P_rated = P_rated

            ## REGION II.5 and RATED with peak shaving##
            if peak_thrust_shaving:
                max_T = ps_percent * T_rated

                def const_Urated_Tpeak(x):
                    pitch_i = x[0]
                    Uhub_i = x[1]
                    Omega_i = min([Uhub_i * tsr / Rtip_cone, Omega_max])
                    Omega_i_rpm = Omega_i * 30.0 / np.pi
                    myout, _ = self.ccblade.evaluate([Uhub_i], [Omega_i_rpm], [pitch_i], coefficients=False)
                    P_aero_i = float(myout["P"][0])
                    # P_i,_  = compute_P_and_eff(P_aero_i.flatten(), P_rated, Omega_i_rpm, driveType, driveEta)
                    eff_i = np.interp(Omega_i_rpm, lss_rpm, driveEta)
                    if isinstance(eff_i, type(np.array([]))): eff_i = eff_i[0]
                    P_i = float(P_aero_i * eff_i)
                    T_i = float(myout["T"][0])
                    return 1e-4 * (P_i - P_rated), 1e-4 * (T_i - max_T)

                # Have to search over both pitch and speed
                x0 = [0.0, U_rated]
                i_Uhub = np.max([0, i - 2])
                bnds = [[min_pitch, 15.0], [Uhub[i_Uhub] + TOL, Uhub[-1] - TOL]]
                const = {}
                const["type"] = "eq"
                const["fun"] = const_Urated_Tpeak
                params_rated = minimize(
                    lambda x: x[1], x0, method="slsqp", bounds=bnds, constraints=const, tol=TOL, options={"maxiter": 20, "disp": False}
                )

                if params_rated.success and not np.isnan(params_rated.x[1]):
                    U_rated = params_rated.x[1]
                    pitch_rated = params_rated.x[0]
                else:
                    U_rated = U_rated  # Use guessed value earlier
                    pitch_rated = min_pitch

                Omega_tsr_rated = U_rated * tsr / Rtip_cone
                Omega_rated = np.minimum(Omega_tsr_rated, Omega_max)
                Omega_rpm_rated = Omega_rated * 30.0 / np.pi
                myout, _ = self.ccblade.evaluate([U_rated], [Omega_rpm_rated], [pitch_rated], coefficients=True)
                (
                    P_aero_rated,
                    T_rated,
                    Q_rated,
                    M_rated,
                    Cp_aero_rated,
                    Ct_aero_rated,
                    Cq_aero_rated,
                    Cm_aero_rated,
                ) = [float(myout[key][0]) for key in ["P", "T", "Q", "Mb", "CP", "CT", "CQ", "CMb"]]
                eff_rated = np.interp(Omega_rpm_rated, lss_rpm, driveEta)
                Cp_rated = Cp_aero_rated * eff_rated
                P_rated = P_rated

        else:
            # No rated conditions, so just assume last values
            U_rated = Uhub[-1] + 1e-6
            Omega_tsr_rated = Omega_tsr[-1]
            Omega_rated = Omega[-1]
            Omega_rpm_rated = Omega_rpm[-1]
            pitch_rated = pitch[-1]
            P_aero_rated = P_aero[-1]
            P_rated = P[-1]
            T_rated = T[-1]
            Q_rated = Q[-1]
            M_rated = M[-1]
            Cp_rated = Cp[-1]
            Cp_aero_rated = Cp_aero[-1]
            Ct_aero_rated = Ct_aero[-1]
            Cq_aero_rated = Cq_aero[-1]
            Cm_aero_rated = Cm_aero[-1]
            eff_rated = eff[-1]

        # Store rated speed in array
        Uhub = np.r_[Uhub, U_rated]
        isort = np.argsort(Uhub)
        Uhub = Uhub[isort]

        Omega_tsr = np.r_[Omega_tsr, Omega_tsr_rated][isort]
        Omega = np.r_[Omega, Omega_rated][isort]
        Omega_rpm = np.r_[Omega_rpm, Omega_rpm_rated][isort]
        pitch = np.r_[pitch, pitch_rated][isort]
        P_aero = np.r_[P_aero, P_aero_rated][isort]
        P = np.r_[P, P_rated][isort]
        T = np.r_[T, T_rated][isort]
        Q = np.r_[Q, Q_rated][isort]
        M = np.r_[M, M_rated][isort]
        Cp = np.r_[Cp, Cp_rated][isort]
        Cp_aero = np.r_[Cp_aero, Cp_aero_rated][isort]
        Ct_aero = np.r_[Ct_aero, Ct_aero_rated][isort]
        Cq_aero = np.r_[Cq_aero, Cq_aero_rated][isort]
        Cm_aero = np.r_[Cm_aero, Cm_aero_rated][isort]
        eff = np.r_[eff, eff_rated][isort]

        i_rated = np.where(Uhub == U_rated)[0][0]
        i_3 = np.minimum(i_rated + 1, self.n_pc)

        # Set rated conditions for rest of powercurve
        Omega[i_rated:] = Omega_rated  # Stay at this speed if hit rated too early
        Omega_rpm = Omega * 30.0 / np.pi

        ## REGION II ##
        # Functions to be used inside of power maximization until Region 3
        def maximizePower(pitch_i, Uhub_i, Omega_rpm_i, scaling_power):
            myout, _ = self.ccblade.evaluate([Uhub_i], [Omega_rpm_i], [pitch_i], coefficients=False)
            return -myout["P"] / scaling_power

        def constr_Tmax(pitch_i, Uhub_i, Omega_rpm_i, scaling_thrust):
            myout, _ = self.ccblade.evaluate([Uhub_i], [Omega_rpm_i], [pitch_i], coefficients=False)
            return (max_T - float(myout["T"][0])) / scaling_thrust

        # Maximize power until rated
        for i in range(i_3):
            # No need to optimize if already doing well or if flag
            # fix_pitch_regI12, which locks pitch in region I 1/2, is on
            # For the 1.04 value on thrust shaving, that is about the level of compliance we see for the Region 2.5
            # points coming out of the optimization routines in the next block of code. This way, a Region 2 point that is
            # at optimal TSR-rpm and 0-deg pitch but 1% over the thrust target can be allowed to stand.
            if (
                ((Omega[i] == Omega_tsr[i]) and not peak_thrust_shaving)
                or ((Omega[i] == Omega_tsr[i]) and peak_thrust_shaving and (T[i]/max_T <= 1.04))
                or ((Omega[i] == Omega_min) and discrete_inputs["fix_pitch_regI12"])
                or (found_rated and (i == i_rated))
            ):
                continue

            # Find pitch value that gives highest power rating
            pitch0 = pitch[i] if i == 0 else pitch[i - 1]
            bnds = [max([min_pitch, pitch0 - 10.0]), max([min_pitch, pitch0 + 10.0])]
            # For a successfull minimization, find the initial power value to nondimensionalize power and bring the figure of merit close to 1
            myout, _ = self.ccblade.evaluate(Uhub[i], Omega_rpm[i], pitch0, coefficients=False)
            # For better conditioning near cut-in, use rated values
            scaling_power = 0.1*P_rated #myout["P"]
            scaling_thrust = 0.1*T_rated #myout["T"]
            if peak_thrust_shaving and found_rated and (Omega[i] <= Omega_tsr[i]):
                # Have to constrain thrust in Region II.5
                const = {}
                const["type"] = "ineq"
                const["fun"] = lambda x: constr_Tmax(x, Uhub[i], Omega_rpm[i], scaling_thrust)
                params = minimize(
                    lambda x: maximizePower(x, Uhub[i], Omega_rpm[i], scaling_power),
                    pitch0,
                    method="slsqp", #"cobyla",
                    bounds=[bnds],
                    constraints=const,
                    tol=TOL,
                    options={"maxiter": 20, "disp": False},  #'catol':0.01*max_T},
                )
                pitch[i] = params.x[0]
            else:
                # Only adjust pitch- should be mostly region I.5 for peak shaving and also II.5 for non-peak shaving
                pitch[i] = minimize_scalar(
                    lambda x: maximizePower(x, Uhub[i], Omega_rpm[i], scaling_power),
                    bounds=bnds,
                    method="bounded",
                    options={"disp": False, "xatol": TOL, "maxiter": 40},
                )["x"][0]
                
            # Find associated power
            myout, _ = self.ccblade.evaluate([Uhub[i]], [Omega_rpm[i]], [pitch[i]], coefficients=True)
            P_aero[i], T[i], Q[i], M[i], Cp_aero[i], Ct_aero[i], Cq_aero[i], Cm_aero[i] = [
                myout[key][0] for key in ["P", "T", "Q", "Mb", "CP", "CT", "CQ", "CMb"]
            ]
            # P[i], eff[i] = compute_P_and_eff(P_aero[i], P_rated, Omega_rpm[i], driveType, driveEta)
            eff[i] = np.interp(Omega_rpm[i], lss_rpm, driveEta)
            P[i] = P_aero[i] * eff[i]
            Cp[i] = Cp_aero[i] * eff[i]

        ## REGION III ##
        # JPJ: this part can be converted into a BalanceComp with a solver.
        # This will be less expensive and allow us to get derivatives through the process.
        if region3:
            # Function to be used to stay at rated power in Region 3
            def rated_power_dist(pitch_i, Uhub_i, Omega_rpm_i):
                myout, _ = self.ccblade.evaluate([Uhub_i], [Omega_rpm_i], [pitch_i], coefficients=False)
                P_aero_i = myout["P"]
                eff_i = np.interp(Omega_rpm_i, lss_rpm, driveEta)
                P_i = P_aero_i * eff_i
                return 1e-4 * (P_i[0] - P_rated)

            # Solve for Region 3 pitch
            if self.regulation_reg_III:
                for i in range(i_3, self.n_pc):
                    pitch0 = pitch[i - 1]
                    bnds = ([pitch0, pitch0 + 15.0],)
                    try:
                        pitch[i] = brentq(
                            lambda x: rated_power_dist(x, Uhub[i], Omega_rpm[i]),
                            bnds[0][0],
                            bnds[0][1],
                            xtol=1e-1 * TOL,
                            rtol=1e-2 * TOL,
                            maxiter=40,
                            disp=False,
                        )
                    except ValueError:
                        pitch[i] = minimize_scalar(
                            lambda x: np.abs(rated_power_dist(x, Uhub[i], Omega_rpm[i])),
                            bounds=bnds[0],
                            method="bounded",
                            options={"disp": False, "xatol": TOL, "maxiter": 40},
                        )["x"]

                    myout, _ = self.ccblade.evaluate([Uhub[i]], [Omega_rpm[i]], [pitch[i]], coefficients=True)
                    P_aero[i], T[i], Q[i], M[i], Cp_aero[i], Ct_aero[i], Cq_aero[i], Cm_aero[i] = [
                        myout[key][0] for key in ["P", "T", "Q", "Mb", "CP", "CT", "CQ", "CMb"]
                    ]
                    eff[i] = np.interp(Omega_rpm[i], lss_rpm, driveEta)
                    P[i] = P_aero[i] * eff[i]
                    Cp[i] = Cp_aero[i] * eff[i]
                    # P[i]        = P_rated

                    # If we are thrust shaving, then check if this is a point that must be modified
                    if peak_thrust_shaving and T[i] >= max_T:
                        myout, _ = self.ccblade.evaluate(Uhub[i], Omega_rpm[i], pitch0, coefficients=False)
                        scaling_thrust = 0.1*myout["T"]
                        const = {}
                        const["type"] = "ineq"
                        const["fun"] = lambda x: constr_Tmax(x, Uhub[i], Omega_rpm[i], scaling_thrust)
                        params = minimize(
                            lambda x: np.abs(rated_power_dist(x, Uhub[i], Omega_rpm[i])),
                            pitch0,
                            method="slsqp",
                            bounds=bnds,
                            constraints=const,
                            tol=TOL,
                            options={"maxiter": 20, "disp": False},
                        )
                        if params.success and not np.isnan(params.x[0]):
                            pitch[i] = params.x[0]

                        myout, _ = self.ccblade.evaluate([Uhub[i]], [Omega_rpm[i]], [pitch[i]], coefficients=True)
                        P_aero[i], T[i], Q[i], M[i], Cp_aero[i], Ct_aero[i], Cq_aero[i], Cm_aero[i] = [
                            myout[key] for key in ["P", "T", "Q", "Mb", "CP", "CT", "CQ", "CMb"]
                        ]
                        eff[i] = np.interp(Omega_rpm[i], lss_rpm, driveEta)
                        P[i] = P_aero[i] * eff[i]
                        Cp[i] = Cp_aero[i] * eff[i]
                        # P[i]        = P_rated

            else:
                P[i_3:] = P_rated
                P_aero[i_3:] = P_aero[i_3 - 1]
                T[i_3:] = 0
                Q[i_3:] = P[i_3:] / Omega[i_3:]
                M[i_3:] = 0
                pitch[i_3:] = 0
                Cp[i_3:] = P[i_3:] / (0.5 * inputs["rho"][0]* np.pi * Rtip_cone**2 * Uhub[i_3:] ** 3)
                Cp_aero[i_3:] = P_aero[i_3:] / (0.5 * inputs["rho"][0] * np.pi * Rtip_cone**2 * Uhub[i_3:] ** 3)
                Ct_aero[i_3:] = 0
                Cq_aero[i_3:] = 0
                Cm_aero[i_3:] = 0

        ## END POWERCURVE ##

        # Store outputs
        outputs["T"] = T
        outputs["Q"] = Q
        outputs["Omega"] = Omega_rpm

        outputs["P"] = P
        outputs["Cp"] = Cp
        outputs["P_aero"] = P_aero
        outputs["Cp_aero"] = Cp_aero
        outputs["Ct_aero"] = Ct_aero
        outputs["Cq_aero"] = Cq_aero
        outputs["Cm_aero"] = Cm_aero
        outputs["V"] = Uhub
        outputs["M"] = M
        outputs["pitch"] = pitch

        outputs["rated_V"] = np.float64(U_rated)
        outputs["rated_Omega"] = Omega_rpm_rated
        outputs["rated_pitch"] = pitch_rated
        outputs["rated_T"] = T_rated
        outputs["rated_Q"] = Q_rated
        outputs["rated_mech"] = P_aero_rated
        outputs["rated_efficiency"] = eff_rated

        self.ccblade.induction_inflow = True
        tsr_vec = Omega_rpm / 30.0 * np.pi * Rtip_cone / Uhub
        id_regII = np.argmin(abs(tsr_vec - inputs["tsr_operational"][0]))
        ax_induct_rotor = np.zeros_like(Uhub)
        for i in range(len(Uhub)):
            loads, _ = self.ccblade.distributedAeroLoads(Uhub[i], Omega_rpm[i], pitch[i], 0.0)
            ax_induct_rotor[i] = 2. / inputs["r"][-1]**2. * np.trapz(loads['a'] * inputs["r"], inputs["r"])
            if i == id_regII:
                # outputs
                outputs["ax_induct_regII"] = loads["a"]
                outputs["tang_induct_regII"] = loads["ap"]
                outputs["aoa_regII"] = loads["alpha"]
                outputs["cl_regII"] = loads["Cl"]
                outputs["cd_regII"] = loads["Cd"]
                outputs["L_D"] = loads["Cl"] / loads["Cd"]
                outputs["Cp_regII"] = Cp_aero[i]
                outputs["Ct_regII"] = Ct_aero[i]
        outputs["ax_induct_rotor"] = ax_induct_rotor

class ComputeSplines(ExplicitComponent):
    """
    Compute splined quantities for V, P, and Omega.
    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        modeling_options = self.options["modeling_options"]
        self.n_pc = modeling_options["WISDEM"]["RotorSE"]["n_pc"]
        self.n_pc_spline = modeling_options["WISDEM"]["RotorSE"]["n_pc_spline"]

        self.add_input("v_min", val=0.0, units="m/s", desc="cut-in wind speed")
        self.add_input("v_max", val=0.0, units="m/s", desc="cut-out wind speed")
        self.add_input("V", val=np.zeros(self.n_pc), units="m/s", desc="wind vector")
        self.add_input("Omega", val=np.zeros(self.n_pc), units="rpm", desc="rotor rotational speed")
        self.add_input("P", val=np.zeros(self.n_pc), units="W", desc="rotor electrical power")

        self.add_output("V_spline", val=np.zeros(self.n_pc_spline), units="m/s", desc="wind vector")
        self.add_output("P_spline", val=np.zeros(self.n_pc_spline), units="W", desc="rotor electrical power")
        self.add_output("Omega_spline", val=np.zeros(self.n_pc_spline), units="rpm", desc="omega")

        self.declare_partials(of="V_spline", wrt="v_min")
        self.declare_partials(of="V_spline", wrt="v_max")

        self.declare_partials(of="P_spline", wrt="v_min", method="fd")
        self.declare_partials(of="P_spline", wrt="v_max", method="fd")
        self.declare_partials(of="P_spline", wrt="V", method="fd")
        self.declare_partials(of="P_spline", wrt="P", method="fd")

        self.declare_partials(of="Omega_spline", wrt="v_min", method="fd")
        self.declare_partials(of="Omega_spline", wrt="v_max", method="fd")
        self.declare_partials(of="Omega_spline", wrt="V", method="fd")
        self.declare_partials(of="Omega_spline", wrt="Omega", method="fd")

    def compute(self, inputs, outputs):
        # Fit spline to powercurve for higher grid density
        V_spline = np.linspace(inputs["v_min"][0], inputs["v_max"][0], self.n_pc_spline)
        spline = PchipInterpolator(inputs["V"], inputs["P"])
        P_spline = spline(V_spline)
        spline = PchipInterpolator(inputs["V"], inputs["Omega"])
        Omega_spline = spline(V_spline)

        # outputs
        outputs["V_spline"] = V_spline.flatten()
        outputs["P_spline"] = P_spline.flatten()
        outputs["Omega_spline"] = Omega_spline.flatten()

    def compute_partials(self, inputs, partials):
        linspace_with_deriv
        V_spline, dy_dstart, dy_dstop = linspace_with_deriv(inputs["v_min"][0], inputs["v_max"][0], self.n_pc_spline)
        partials["V_spline", "v_min"] = dy_dstart
        partials["V_spline", "v_max"] = dy_dstop


# Class to define a constraint so that the blade cannot operate in stall conditions
class NoStallConstraint(ExplicitComponent):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        modeling_options = self.options["modeling_options"]
        self.n_span = n_span = modeling_options["WISDEM"]["RotorSE"]["n_span"]
        self.n_aoa = n_aoa = modeling_options["WISDEM"]["RotorSE"]["n_aoa"]  # Number of angle of attacks
        self.n_Re = n_Re = modeling_options["WISDEM"]["RotorSE"]["n_Re"]  # Number of Reynolds, so far hard set at 1
        self.n_tab = n_tab = modeling_options["WISDEM"]["RotorSE"][
            "n_tab"
        ]  # Number of tabulated data. For distributed aerodynamic control this could be > 1

        self.add_input(
            "s",
            val=np.zeros(n_span),
            desc="1D array of the non-dimensional spanwise grid defined along blade axis (0-blade root, 1-blade tip)",
        )
        self.add_input("aoa_along_span", val=np.zeros(n_span), units="deg", desc="Angle of attack along blade span")
        self.add_input("stall_margin", val=3.0, units="deg", desc="Minimum margin from the stall angle")
        self.add_input(
            "min_s",
            val=0.25,
            desc="Minimum nondimensional coordinate along blade span where to define the constraint (blade root typically stalls)",
        )
        self.add_input("airfoils_cl", val=np.zeros((n_span, n_aoa, n_Re, n_tab)), desc="lift coefficients, spanwise")
        self.add_input("airfoils_cd", val=np.zeros((n_span, n_aoa, n_Re, n_tab)), desc="drag coefficients, spanwise")
        self.add_input("airfoils_cm", val=np.zeros((n_span, n_aoa, n_Re, n_tab)), desc="moment coefficients, spanwise")
        self.add_input("airfoils_aoa", val=np.zeros((n_aoa)), units="deg", desc="angle of attack grid for polars")

        self.add_output(
            "no_stall_constraint",
            val=np.zeros(n_span),
            desc="Constraint, ratio between angle of attack plus a margin and stall angle",
        )
        self.add_output(
            "stall_angle_along_span", val=np.zeros(n_span), units="deg", desc="Stall angle along blade span"
        )

    def compute(self, inputs, outputs):
        i_min = np.argmin(np.abs(inputs["min_s"] - inputs["s"]))
        n_span = len(inputs["s"])

        for i in range(n_span):
            unsteady = eval_unsteady(
                inputs["airfoils_aoa"],
                inputs["airfoils_cl"][i, :, 0, 0],
                inputs["airfoils_cd"][i, :, 0, 0],
                inputs["airfoils_cm"][i, :, 0, 0],
            )
            outputs["stall_angle_along_span"][i] = unsteady["alpha1"]
            if outputs["stall_angle_along_span"][i] == 0:
                outputs["stall_angle_along_span"][i] = 1e-6  # To avoid nan

        for i in range(i_min, n_span):
            outputs["no_stall_constraint"][i] = (inputs["aoa_along_span"][i] + inputs["stall_margin"][0]) / outputs[
                "stall_angle_along_span"
            ][i]

            if outputs["stall_angle_along_span"][i] <= 1.0e-6:
                outputs["no_stall_constraint"][i] = 0.0

            logger.debug(
                "Blade is violating the minimum margin to stall at span location %.2f %%" % (inputs["s"][i] * 100.0)
            )


class AEP(ExplicitComponent):
    def initialize(self):
        self.options.declare("nspline")

    def setup(self):
        n_pc_spline = self.options["nspline"]
        """integrate to find annual energy production"""

        # inputs
        self.add_input(
            "CDF_V",
            val=np.zeros(n_pc_spline),
            units="m/s",
            desc="cumulative distribution function evaluated at each wind speed",
        )
        self.add_input("P", val=np.zeros(n_pc_spline), units="W", desc="power curve (power)")
        self.add_input(
            "lossFactor", val=1.0, desc="multiplicative factor for availability and other losses (soiling, array, etc.)"
        )

        # outputs
        self.add_output("AEP", val=0.0, units="kW*h", desc="annual energy production")

        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

    def compute(self, inputs, outputs):
        lossFactor = inputs["lossFactor"][0]
        P = inputs["P"]
        CDF_V = inputs["CDF_V"]

        factor = lossFactor / 1e3 * 365.0 * 24.0
        outputs["AEP"] = factor * np.trapz(P, CDF_V)  # in kWh
        """
        dAEP_dP, dAEP_dCDF = trapz_deriv(P, CDF_V)
        dAEP_dP *= factor
        dAEP_dCDF *= factor

        dAEP_dlossFactor = np.array([outputs['AEP']/lossFactor])
        self.J = {}
        self.J['AEP', 'CDF_V'] = np.reshape(dAEP_dCDF, (1, len(dAEP_dCDF)))
        self.J['AEP', 'P'] = np.reshape(dAEP_dP, (1, len(dAEP_dP)))
        self.J['AEP', 'lossFactor'] = dAEP_dlossFactor

    def compute_partials(self, inputs, J):
        J = self.J
        """


def compute_P_and_eff(aeroPower, ratedPower, Omega_rpm, drivetrainType, drivetrainEff):
    if not np.any(drivetrainEff):
        drivetrainType = drivetrainType.upper()
        if drivetrainType == "GEARED":
            constant = 0.01289
            linear = 0.08510
            quadratic = 0.0

        elif drivetrainType == "SINGLE_STAGE":
            constant = 0.01331
            linear = 0.03655
            quadratic = 0.06107

        elif drivetrainType == "MULTI_DRIVE":
            constant = 0.01547
            linear = 0.04463
            quadratic = 0.05790

        elif drivetrainType in ["PM_DIRECT_DRIVE", "DIRECT_DRIVE", "DIRECT DRIVE"]:
            constant = 0.01007
            linear = 0.02000
            quadratic = 0.06899

        elif drivetrainType == "CONSTANT_EFF":
            constant = 0.00
            linear = 0.07
            quadratic = 0.0
        else:
            raise ValueError("The drivetrain model is not supported! Please check rotor_power.py")

        Pbar0 = aeroPower / ratedPower

        # handle negative power case (with absolute value)
        Pbar1, dPbar1_dPbar0 = smooth_abs(Pbar0, dx=0.01)

        # truncate idealized power curve for purposes of efficiency calculation
        Pbar, dPbar_dPbar1, _ = smooth_min(Pbar1, 1.0, pct_offset=0.01)

        # compute efficiency
        eff = 1.0 - (constant / Pbar + linear + quadratic * Pbar)
        eff = np.maximum(eff, 1e-3)
    else:
        # Use table lookup from rpm to calculate total efficiency
        eff = np.interp(Omega_rpm, drivetrainEff[:, 0], drivetrainEff[:, 1])

    return aeroPower * eff, eff


def eval_unsteady(alpha, cl, cd, cm):
    # calculate unsteady coefficients from polars for OpenFAST's Aerodyn

    unsteady = {}
    Re = 1e6  # Does not factor into any calculations
    mypolar = Polar(Re=Re, alpha=alpha, cl=cl, cd=cd, cm=cm, compute_params=True, radians=False)
    (alpha0, alpha1, alpha2, cnSlope, cn1, cn2, cd0, cm0) = mypolar.unsteadyParams()
    unsteady["alpha0"] = alpha0
    unsteady["alpha1"] = alpha1
    unsteady["alpha2"] = alpha2
    unsteady["Cd0"] = cd0
    unsteady["Cm0"] = cm0
    unsteady["Cn1"] = cn1
    unsteady["Cn2"] = cn2
    unsteady["C_nalpha"] = cnSlope
    unsteady["eta_e"] = 1
    unsteady["T_f0"] = "Default"
    unsteady["T_V0"] = "Default"
    unsteady["T_p"] = "Default"
    unsteady["T_VL"] = "Default"
    unsteady["b1"] = "Default"
    unsteady["b2"] = "Default"
    unsteady["b5"] = "Default"
    unsteady["A1"] = "Default"
    unsteady["A2"] = "Default"
    unsteady["A5"] = "Default"
    unsteady["S1"] = 0
    unsteady["S2"] = 0
    unsteady["S3"] = 0
    unsteady["S4"] = 0
    unsteady["St_sh"] = "Default"
    unsteady["k0"] = 0
    unsteady["k1"] = 0
    unsteady["k2"] = 0
    unsteady["k3"] = 0
    unsteady["k1_hat"] = 0
    unsteady["x_cp_bar"] = "Default"
    unsteady["UACutout"] = "Default"
    unsteady["filtCutOff"] = "Default"
    unsteady["Alpha"] = alpha
    unsteady["Cl"] = cl
    unsteady["Cd"] = cd
    unsteady["Cm"] = cm

    return unsteady



from unittest import TestCase
import pandas as pd
from wisdem.landbosse.model import DevelopmentCost


class TestDevelopmentCost(TestCase):
    def setUp(self):
        """
         This setUp() method executes before each test. It creates an
         instance attribute, self.instance, that refers to a DevelopmentCost
         instance under test. input_dict is the dictionary of inputs values
         that configures that instance.
         This is just a default dictionary. Methods that test behavior of
         specific variables customize the relevant key/value pairs.
         The key site_facility_building_area_df is set to None here. The
         dataframe is read from the csv in the site_facility tests below.
        """

        self.input_dict = dict()
        self.input_dict['development_materials_cost_usd'] = 10000
        self.input_dict['development_labor_cost_usd'] = 10000
        self.input_dict['development_equipment_cost_usd'] = 10000
        self.input_dict['development_mobilization_cost_usd'] = 10000
        self.input_dict['development_other_cost_usd'] = 10000
        self.project_name = 'project 1'
        self.output_dict = dict()


    def test_DevelopmentCostModule(self):
        """
        Black box test to check whether module is ran successfully or not
        """
        run_DevelopmentCost = DevelopmentCost(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        trial_run = run_DevelopmentCost.run_module()


        if trial_run[0] == 0 :
            print('\n\n================== MODULE EXECUTION SUCCESS =========================\n')
            print(' DevelopmentCost module ran successfully. See the list of inputs'
                  '\n and outputs below used by the module in its calculations:')
            print( '\n=====================================================================\n')

        elif trial_run[0] == 1 :
            print('\n\n====================================== MODULE EXECUTION FAILURE ======================================\n')
            print(' > DevelopmentCost module failed to run successfully. Error detected: ',trial_run[1] ,
                  '\n > Scroll below to see detailed information about error encountered.'
                  '\n > See the list of inputs below used by the module in its calculations:')
            print('\n========================================================================================================\n')


        print('\nGiven below is the set of inputs fed into DevelopmentCost module:\n')
        for key, value in self.input_dict.items():
            if isinstance(value, pd.DataFrame):
                print(key, ':\n', value)
            else:
                print(key, ':', value)

        if trial_run[0] == 0:  # Only print outputs if module ran successfully.
            print('\nGiven below is the set of outputs calculated by the DevelopmentCost module:\n')
            for key, value in self.output_dict.items():
                if isinstance(value, pd.DataFrame):
                    print('\nNow printing DataFrame ->', key, ':\n', value)
                else:
                    print(key, ':', value)


import os

import openmdao.api as om

import wisdem.inputs as sch
from wisdem.glue_code.gc_RunTools import PlotRecorder

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
fname_analysis_options = mydir + os.sep + "analysis_options.yaml"
analysis_options = sch.load_analysis_yaml(fname_analysis_options)
wt_opt = om.Problem(reports=False, model=PlotRecorder(opt_options=analysis_options))
wt_opt.setup(derivatives=False)
wt_opt.run_model()



import openmdao.api as om


class IntermittentComponent(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("num_iterations_between_calls", 3)
        self.frozen_outputs = {}
        self.actual_compute_calls = 0

    def compute(self, inputs, outputs):
        """
        This is a wrapper for the actual compute call, `internal_compute()`,
        which calls the compute method every `num_iterations_between_calls`.
        This allows more expensive analyses to be run less often at the expense
        of accuracy in an optimization context.
        If you want the compute to be run only once at the beginning of the
        optimization, set `num_iterations_between_calls` to a very large number.
        """

        num_iterations_between_calls = self.options["num_iterations_between_calls"]

        # Determine if we are in a compute call in which we want to update the
        # frozen outputs
        regular_compute = (self.iter_count_without_approx % num_iterations_between_calls) == 0 and not self.under_approx
        approx_compute = (
            (self.iter_count_without_approx - 1) % num_iterations_between_calls
        ) == 0 and self.under_approx

        # If we're within one of those types of compute calls, call the actual
        # internal_compute() method to update the results
        if regular_compute or approx_compute:
            self.internal_compute(inputs, outputs)
            self.actual_compute_calls += 1

            # Save off the results to the frozen_outputs dict
            for key in outputs:
                self.frozen_outputs[key] = outputs[key]

        # If we're using the frozen results, simply set the outputs from those results
        else:
            for key in outputs:
                outputs[key] = self.frozen_outputs[key]

    def internal_compute(self, inputs, outputs):
        """
        This is the actual method where the computations should occur.
        """
        raise NotImplementedError("Please define an `internal_compute` method.")



import numpy as np
import pandas as pd
import wisdem.commonse.utilities as util


def is_floating(prob):
    return prob.model.options["modeling_options"]["flags"]["floating"]


def is_monopile(prob):
    return prob.model.options["modeling_options"]["flags"]["monopile"]


def get_tower_diameter(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.monopile_outer_diameter"], prob["towerse.tower_outer_diameter"][1:]]
    else:
        return prob["towerse.tower_outer_diameter"]


def get_tower_thickness(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.monopile_wall_thickness"], prob["towerse.tower_wall_thickness"]]
    else:
        return prob["towerse.tower_wall_thickness"]


def get_zpts(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.z_param"], prob["towerse.z_param"][1:]]
    else:
        return prob["towerse.z_param"]


def get_section_height(prob):
    return np.diff(get_zpts(prob))


def get_transition_height(prob):
    return prob["towerse.foundation_height"]


def get_tower_outfitting(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.member.outfitting_factor"], prob["towerse.member.outfitting_factor"][1:]]
    else:
        return prob["towerse.member.outfitting_factor"]


def get_tower_E(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.member.E"], prob["towerse.member.E"][1:]]
    else:
        return prob["towerse.member.E"]


def get_tower_G(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.member.G"], prob["towerse.member.G"][1:]]
    else:
        return prob["towerse.member.G"]


def get_tower_rho(prob):
    if is_monopile(prob):
        return np.r_[prob["fixedse.member.rho"], prob["towerse.member.rho"][1:]]
    else:
        return prob["towerse.member.rho"]


def get_tower_mass(prob):
    return prob["towerse.tower_mass"]


def get_tower_cost(prob):
    return prob["towerse.tower_cost"]


def get_monopile_mass(prob):
    return prob["fixedse.monopile_mass"]


def get_monopile_cost(prob):
    return prob["fixedse.monopile_cost"]


def get_structural_mass(prob):
    if is_monopile(prob):
        return prob["fixedse.structural_mass"]
    else:
        return prob["towerse.tower_mass"]


def get_structural_cost(prob):
    if is_monopile(prob):
        return prob["fixedse.structural_cost"]
    else:
        return prob["towerse.tower_cost"]


def get_tower_freqs(prob):
    if is_monopile(prob):
        return prob["fixedse.structural_frequencies"]
    else:
        return prob["towerse.tower.structural_frequencies"]


def get_tower_cm(prob):
    return prob["towerse.tower_center_of_mass"]


def get_tower_cg(prob):
    return get_tower_cm(prob)


def get_blade_shape(prob):
    blade_shape  = np.c_[prob.get_val('blade.outer_shape_bem.s'),
                         prob.get_val('blade.outer_shape_bem.ref_axis','m')[:,2],
                         prob.get_val('blade.outer_shape_bem.chord','m'),
                         prob.get_val('blade.outer_shape_bem.twist', 'deg'),
                         prob.get_val('blade.interp_airfoils.r_thick_interp')*100,
                         prob.get_val('blade.outer_shape_bem.pitch_axis')*100,
                         prob.get_val('blade.outer_shape_bem.ref_axis','m')[:,0],
                         prob.get_val('blade.outer_shape_bem.ref_axis','m')[:,1],
                         ]
    blade_shape_col = ['Blade Span','Rotor Coordinate [m]',
                       'Chord [m]', 'Twist [deg]',
                       'Relative Thickness [%]', 'Pitch Axis Chord Location [%]',
                       'Prebend [m]', 'Sweep [m]']
    return pd.DataFrame(data=blade_shape, columns=blade_shape_col)


def get_blade_elasticity(prob):
    blade_stiff  = np.c_[prob.get_val('rotorse.r','m'),
                         prob.get_val('rotorse.A','m**2'),
                         prob.get_val('rotorse.EA','N'),
                         prob.get_val('rotorse.EIxx','N*m**2'),
                         prob.get_val('rotorse.EIyy','N*m**2'),
                         prob.get_val('rotorse.EIxy','N*m**2'),
                         prob.get_val('rotorse.GJ','N*m**2'),
                         prob.get_val('rotorse.rhoA','kg/m'),
                         prob.get_val('rotorse.rhoJ','kg*m'),
                         prob.get_val('rotorse.re.x_tc','mm'),
                         prob.get_val('rotorse.re.y_tc','mm'),
                         prob.get_val('rotorse.re.precomp.x_sc','mm'),
                         prob.get_val('rotorse.re.precomp.y_sc','mm'),
                         prob.get_val('rotorse.re.x_cg','mm'),
                         prob.get_val('rotorse.re.y_cg','mm'),
                         prob.get_val('rotorse.re.flap_iner','kg/m'),
                         prob.get_val('rotorse.re.edge_iner','kg/m')]
    blade_stiff_col = ['Blade Span [m]',
                       'Cross-sectional area [m^2]',
                       'Axial stiffness [N]',
                       'Edgewise stiffness [Nm^2]',
                       'Flapwise stiffness [Nm^2]',
                       'Flap-edge coupled stiffness [Nm^2]',
                       'Torsional stiffness [Nm^2]',
                       'Mass density [kg/m]',
                       'Polar moment of inertia density [kg*m]',
                       'X-distance to tension center [mm]',
                       'Y-distance to tension center [mm]',
                       'X-distance to shear center [mm]',
                       'Y-distance to shear center [mm]',
                       'X-distance to mass center [mm]',
                       'Y-distance to mass center [mm]',
                       'Section flap inertia [kg/m]',
                       'Section edge inertia [kg/m]',
                       ]
    return pd.DataFrame(data=blade_stiff, columns=blade_stiff_col)


def get_rotor_performance(prob):
    rotor_perf = np.c_[prob.get_val("rotorse.rp.powercurve.V",'m/s'),
                       prob.get_val("rotorse.rp.powercurve.pitch",'deg'),
                       prob.get_val("rotorse.rp.powercurve.P",'MW'),
                       prob.get_val("rotorse.rp.powercurve.Cp"),
                       prob.get_val("rotorse.rp.powercurve.Cp_aero"),
                       prob.get_val("rotorse.rp.powercurve.Omega",'rpm'),
                       prob.get_val("rotorse.rp.powercurve.Omega",'rad/s')*0.5*prob["configuration.rotor_diameter_user"],
                       prob.get_val("rotorse.rp.powercurve.T",'MN'),
                       prob.get_val("rotorse.rp.powercurve.Ct_aero"),
                       prob.get_val("rotorse.rp.powercurve.Q",'MN*m'),
                       prob.get_val("rotorse.rp.powercurve.Cq_aero"),
                       prob.get_val("rotorse.rp.powercurve.M",'MN*m'),
                       prob.get_val("rotorse.rp.powercurve.Cm_aero"),
                       ]
    rotor_perf_col = ['Wind [m/s]','Pitch [deg]',
                      'Power [MW]','Power Coefficient [-]','Aero Power Coefficient [-]',
                      'Rotor Speed [rpm]','Tip Speed [m/s]',
                      'Thrust [MN]','Thrust Coefficient [-]',
                      'Torque [MNm]','Torque Coefficient [-]',
                      'Blade Moment [MNm]','Blade Moment Coefficient [-]',
                      ]
    return pd.DataFrame(data=rotor_perf, columns=rotor_perf_col)


def get_nacelle_mass(prob):

    # Nacelle mass properties tabular
    # Columns are ['Mass', 'CoM_x', 'CoM_y', 'CoM_z',
    #              'MoI_cm_xx', 'MoI_cm_yy', 'MoI_cm_zz', 'MoI_cm_xy', 'MoI_cm_xz', 'MoI_cm_yz',
    #              'MoI_TT_xx', 'MoI_TT_yy', 'MoI_TT_zz', 'MoI_TT_xy', 'MoI_TT_xz', 'MoI_TT_yz']
    nacDF = prob.model.wt.wt_rna.drivese.nac._mass_table
    hub_cm = prob["drivese.hub_system_cm"][0]
    L_drive = prob["drivese.L_drive"][0]
    tilt = prob.get_val('nacelle.uptilt', 'rad')[0]
    shaft0 = prob["drivese.shaft_start"]
    Cup = -1.0
    hub_cm = R = shaft0 + (L_drive + hub_cm) * np.array([Cup * np.cos(tilt), 0.0, np.sin(tilt)])
    hub_mass = prob['drivese.hub_system_mass']
    hub_I = prob["drivese.hub_system_I"]
    hub_I_TT = util.rotateI(hub_I, -Cup * tilt, axis="y")
    hub_I_TT = util.unassembleI( util.assembleI(hub_I_TT) +
                                 hub_mass * (np.dot(R, R) * np.eye(3) - np.outer(R, R)) )
    blades_mass = prob['drivese.blades_mass']
    blades_I = prob["drivese.blades_I"]
    blades_I_TT = util.rotateI(blades_I, -Cup * tilt, axis="y")
    blades_I_TT = util.unassembleI( util.assembleI(blades_I_TT) +
                                    blades_mass * (np.dot(R, R) * np.eye(3) - np.outer(R, R)) )
    rna_mass = prob['drivese.rna_mass']
    rna_cm = R = prob['drivese.rna_cm']
    rna_I_TT = prob['drivese.rna_I_TT']
    rna_I = util.unassembleI( util.assembleI(rna_I_TT) +
                                    rna_mass * (np.dot(R, R) * np.eye(3) + np.outer(R, R)) )
    nacDF.loc['Blades'] = np.r_[blades_mass, hub_cm, blades_I, blades_I_TT].tolist()
    nacDF.loc['Hub_System'] = np.r_[hub_mass, hub_cm, hub_I, hub_I_TT].tolist()
    nacDF.loc['RNA'] = np.r_[rna_mass, rna_cm, rna_I, rna_I_TT].tolist()
    return nacDF


def get_tower_table(prob):

    # Tabular output: Tower
    float_flag = is_floating(prob)
    water_depth = prob['env.water_depth']
    h_trans = get_transition_height(prob)
    htow = get_zpts(prob) #np.cumsum(np.r_[0.0, prob['towerse.tower_section_height']]) + prob['towerse.z_start']
    t = get_tower_thickness(prob)
    towdata = np.c_[htow,
                    get_tower_diameter(prob),
                    np.r_[t[0], t]]
    rowadd = []
    for k in range(towdata.shape[0]):
        if k==0: continue
        if k+1 < towdata.shape[0]:
            rowadd.append([towdata[k,0]+1e-3, towdata[k,1], towdata[k+1,2]])
    towdata = np.vstack((towdata, rowadd))
    towdata[:,-1] *= 1e3
    towdata = np.round( towdata[towdata[:,0].argsort(),], 3)
    colstr = ['Height [m]','OD [m]', 'Thickness [mm]']
    towDF = pd.DataFrame(data=towdata, columns=colstr)
    mycomments = ['']*towdata.shape[0]
    if not float_flag:
        mycomments[0] = 'Monopile start'
        mycomments[np.where(towdata[:,0] == -water_depth)[0][0]] = 'Mud line'
        mycomments[np.where(towdata[:,0] == 0.0)[0][0]] = 'Water line'
    idx_tow = np.where(towdata[:,0] == h_trans)[0][0]
    mycomments[idx_tow] = 'Tower start'
    mycomments[-1] = 'Tower top'
    towDF['Location'] = mycomments
    towDF = towDF[['Location']+colstr]
    A = 0.25*np.pi*(towDF['OD [m]']**2 - (towDF['OD [m]']-2*1e-3*towDF['Thickness [mm]'])**2)
    I = (1/64.)*np.pi*(towDF['OD [m]']**4 - (towDF['OD [m]']-2*1e-3*towDF['Thickness [mm]'])**4)
    outfitting = np.zeros( len(A) )
    if not float_flag:
        outfitting[:idx_tow] = prob['fixedse.outfitting_factor_in']
    outfitting[idx_tow:] = prob['towerse.outfitting_factor_in']
    towDF['Mass Density [kg/m]'] = outfitting * get_tower_rho(prob)[0] * A
    towDF['Fore-aft inertia [kg.m]'] = towDF['Side-side inertia [kg.m]'] = towDF['Mass Density [kg/m]'] * I/A
    towDF['Fore-aft stiffness [N.m^2]'] = towDF['Side-side stiffness [N.m^2]'] = get_tower_E(prob)[0] * I
    towDF['Torsional stiffness [N.m^2]'] = get_tower_G(prob)[0] * 2*I
    towDF['Axial stiffness [N]'] = get_tower_E(prob)[0] * A
    #with open('tow.tbl','w') as f:
    #    towDF.to_latex(f, index=False)
    return towDF



import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt

import wisdem.floatingse.floating_system as sys
from wisdem.commonse import gravity as g
from wisdem.commonse.cylinder_member import NULL, MEMMAX


class TestPlatform(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.outputs = {}
        self.discrete_inputs = {}
        self.discrete_outputs = {}

        self.opt = {}
        self.opt["floating"] = {}
        self.opt["WISDEM"] = {}
        self.opt["WISDEM"]["FloatingSE"] = {}
        self.opt["floating"]["members"] = {}
        self.opt["floating"]["members"]["n_members"] = n_member = 6
        self.opt["floating"]["members"]["outer_shape"] = n_member * ["circular"]
        self.opt["floating"]["rigid_bodies"] = {}
        self.opt["floating"]["rigid_bodies"]["n_bodies"] = 0
        self.opt["WISDEM"]["FloatingSE"]["frame3dd"] = {}
        self.opt["WISDEM"]["FloatingSE"]["frame3dd"]["shear"] = True
        self.opt["WISDEM"]["FloatingSE"]["frame3dd"]["geom"] = True
        self.opt["WISDEM"]["FloatingSE"]["frame3dd"]["tol"] = 1e-8
        self.opt["WISDEM"]["FloatingSE"]["frame3dd"]["modal"] = False
        self.opt["WISDEM"]["FloatingSE"]["gamma_f"] = 1.35  # Safety factor on loads
        self.opt["WISDEM"]["FloatingSE"]["gamma_m"] = 1.3  # Safety factor on materials
        self.opt["WISDEM"]["FloatingSE"]["gamma_n"] = 1.0  # Safety factor on consequence of failure
        self.opt["WISDEM"]["FloatingSE"]["gamma_b"] = 1.1  # Safety factor on buckling
        self.opt["WISDEM"]["FloatingSE"]["gamma_fatigue"] = 1.755  # Not used
        self.opt["mooring"] = {}
        self.opt["mooring"]["n_attach"] = 3

        for k in range(n_member):
            self.inputs[f"member{k}:nodes_xyz"] = NULL * np.ones((MEMMAX, 3))
            self.inputs[f"member{k}:nodes_r"] = NULL * np.ones(MEMMAX)

        for var in [
            "D",
            "t",
            "A",
            "Asx",
            "Asy",
            "Ixx",
            "Iyy",
            "J0",
            "rho",
            "E",
            "G",
            "TorsC",
            "sigma_y",
        ]:
            for k in range(n_member):
                self.inputs[f"member{k}:section_{var}"] = NULL * np.ones(MEMMAX)

        self.inputs["member0:nodes_xyz"][:2, :] = np.array([[0, 0, 0], [1, 0, 0]])
        self.inputs["member1:nodes_xyz"][:2, :] = np.array([[1, 0, 0], [0.5, 1, 0]])
        self.inputs["member2:nodes_xyz"][:2, :] = np.array([[0.5, 1, 0], [0, 0, 0]])
        self.inputs["member3:nodes_xyz"][:2, :] = np.array([[0, 0, 0], [0, 0, 1]])
        self.inputs["member4:nodes_xyz"][:2, :] = np.array([[1, 0, 0], [0, 0, 1]])
        self.inputs["member5:nodes_xyz"][:2, :] = np.array([[0.5, 1, 0], [0, 0, 1]])
        for k in range(n_member):
            L = np.sqrt(np.sum(np.diff(self.inputs[f"member{k}:nodes_xyz"][:2, :], axis=0) ** 2))
            self.inputs[f"member{k}:nodes_r"][:2] = 0.1 * k * np.ones(2)
            self.inputs[f"member{k}:section_D"][:1] = 2.0
            self.inputs[f"member{k}:section_t"][:1] = 0.1
            self.inputs[f"member{k}:section_A"][:1] = 0.5 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_Asx"][:1] = 0.5 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_Asy"][:1] = 0.5 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_Ixx"][:1] = 2 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_Iyy"][:1] = 2 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_J0"][:1] = 2 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_rho"][:1] = 1e3 / (0.5 * k * np.ones(1) + 1) / L
            self.inputs[f"member{k}:section_E"][:1] = 3 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_G"][:1] = 4 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_TorsC"][:1] = 4 * k * np.ones(1) + 1
            self.inputs[f"member{k}:section_sigma_y"][:1] = 5 * k * np.ones(1) + 1
            self.inputs[f"member{k}:idx_cb"] = np.array([ 0 ], dtype=np.int_)
            self.inputs[f"member{k}:buoyancy_force"] = np.array([ 1e2 ])
            self.inputs[f"member{k}:displacement"] = np.array([ 1e1 ])
            self.inputs[f"member{k}:center_of_buoyancy"] = self.inputs[f"member{k}:nodes_xyz"][:2, :].mean(axis=0)
            self.inputs[f"member{k}:center_of_mass"] = self.inputs[f"member{k}:nodes_xyz"][:2, :].mean(axis=0)
            self.inputs[f"member{k}:total_mass"] = np.array([ 1e3 ])
            self.inputs[f"member{k}:total_cost"] = np.array([ 2e3 ])
            self.inputs[f"member{k}:I_total"] = 1e2 + np.arange(6)
            self.inputs[f"member{k}:Awater"] = np.array([ 5.0 ])
            self.inputs[f"member{k}:Iwaterx"] = np.array([ 15.0 ])
            self.inputs[f"member{k}:Iwatery"] = np.array([ 15.0 ])
            self.inputs[f"member{k}:added_mass"] = np.arange(6)
            self.inputs[f"member{k}:ballast_mass"] = np.array([ 1e2 ])
            self.inputs[f"member{k}:variable_ballast_capacity"] = np.array([ 10 + k ])
            self.inputs[f"member{k}:variable_ballast_spts"] = np.linspace(0, 0.5, 10)
            self.inputs[f"member{k}:variable_ballast_Vpts"] = np.arange(10)
            self.inputs[f"member{k}:waterline_centroid"] = self.inputs[f"member{k}:nodes_xyz"][:2, :2].mean(axis=0)

        myones = np.ones(2)

        self.inputs["mooring_neutral_load"] = np.zeros((3, 3))
        self.inputs["mooring_neutral_load"][:, 0] = [200, -100.0, -100]
        self.inputs["mooring_neutral_load"][:, 1] = [0.0, 50, -50]
        self.inputs["mooring_neutral_load"][:, 2] = -1e3
        self.inputs["mooring_fairlead_joints"] = np.array([[0.0, 0.0, 0.0], [0.5, 1.0, 0.0], [1.0, 0.0, 0.0]])
        self.inputs["mooring_stiffness"] = 5 * np.eye(6)
        self.inputs["transition_node"] = self.inputs["member5:nodes_xyz"][1, :]
        self.inputs["turbine_mass"] = np.array([ 1e4 ])
        self.inputs["turbine_cg"] = np.array([0, 0, 50])
        self.inputs["turbine_I"] = 1e6 * np.ones(6)
        self.inputs["transition_piece_mass"] = np.array([ 1e3 ])
        self.inputs["transition_piece_cost"] = np.array([ 3e3 ])
        self.inputs["rho_water"] = np.array([ 1e3 ])

    def testTetrahedron(self):
        myobj = sys.PlatformFrame(options=self.opt)
        myobj.node_mem2glob = {}
        myobj.node_glob2mem = {}
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)

        # Check NULLs and implied number of nodes / elements
        npt.assert_equal(self.outputs["platform_nodes"][4:, :], NULL)
        npt.assert_equal(self.outputs["platform_Fnode"][4:, :], NULL)
        npt.assert_equal(self.outputs["platform_Rnode"][4:], NULL)
        npt.assert_equal(self.outputs["platform_elem_n1"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_n2"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_D"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_t"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_A"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_Asx"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_Asy"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_Ixx"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_Iyy"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_J0"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_rho"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_E"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_G"][6:], NULL)
        npt.assert_equal(self.outputs["platform_elem_sigma_y"][6:], NULL)

        npt.assert_equal(
            self.outputs["platform_nodes"][:4, :],
            np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.5, 1.0, 0.0], [1.0, 0.0, 0.0]]),
        )
        npt.assert_equal(
            self.outputs["platform_Fnode"][:4, :],
            1e2 * np.array([[0.0, 0.0, 2], [0.0, 0.0, 0.0], [0.0, 0.0, 2], [0.0, 0.0, 2.0]]),
        )
        npt.assert_equal(self.outputs["platform_Rnode"][:4], 0.1 * np.r_[3, 5, 5, 4])
        npt.assert_equal(self.outputs["platform_elem_n1"][:6], np.r_[0, 3, 2, 0, 3, 2])
        npt.assert_equal(self.outputs["platform_elem_n2"][:6], np.r_[3, 2, 0, 1, 1, 1])
        npt.assert_equal(self.outputs["platform_elem_D"][:6], 2.0)
        npt.assert_equal(self.outputs["platform_elem_t"][:6], 0.1)
        npt.assert_equal(self.outputs["platform_elem_A"][:6], 0.5 * np.arange(6) + 1)
        npt.assert_equal(self.outputs["platform_elem_Asx"][:6], 0.5 * np.arange(6) + 1)
        npt.assert_equal(self.outputs["platform_elem_Asy"][:6], 0.5 * np.arange(6) + 1)
        npt.assert_equal(self.outputs["platform_elem_Ixx"][:6], 2 * np.arange(6) + 1)
        npt.assert_equal(self.outputs["platform_elem_Iyy"][:6], 2 * np.arange(6) + 1)
        npt.assert_equal(self.outputs["platform_elem_J0"][:6], 2 * np.arange(6) + 1)
        # npt.assert_equal(self.outputs["platform_elem_rho"][:6], 3 * np.arange(6)+1)
        npt.assert_equal(self.outputs["platform_elem_E"][:6], 3 * np.arange(6) + 1)
        npt.assert_equal(self.outputs["platform_elem_G"][:6], 4 * np.arange(6) + 1)
        npt.assert_equal(self.outputs["platform_elem_sigma_y"][:6], 5 * np.arange(6) + 1)
        npt.assert_equal(self.outputs["platform_displacement"], 6e1)
        centroid = np.array([0.375, 0.25, 0.25])
        R = np.zeros((6, 2))
        R[0,] = self.inputs["member0:nodes_xyz"][:2, :2].mean(axis=0) - centroid[:2]
        R[1,] = self.inputs["member1:nodes_xyz"][:2, :2].mean(axis=0) - centroid[:2]
        R[2,] = self.inputs["member2:nodes_xyz"][:2, :2].mean(axis=0) - centroid[:2]
        R[3,] = self.inputs["member3:nodes_xyz"][:2, :2].mean(axis=0) - centroid[:2]
        R[4,] = self.inputs["member4:nodes_xyz"][:2, :2].mean(axis=0) - centroid[:2]
        R[5,] = self.inputs["member5:nodes_xyz"][:2, :2].mean(axis=0) - centroid[:2]

        npt.assert_equal(self.outputs["platform_center_of_buoyancy"], centroid)
        npt.assert_equal(self.outputs["platform_centroid"], centroid)
        cg = (6e3 * centroid + 1e3 * np.array([0.0, 0.0, 1.0])) / 7e3
        npt.assert_equal(self.outputs["platform_hull_center_of_mass"], cg)
        npt.assert_equal(self.outputs["platform_ballast_mass"], 6e2)
        npt.assert_equal(self.outputs["platform_hull_mass"], 6e3 + 1e3 - 6e2)
        npt.assert_equal(self.outputs["platform_cost"], 6 * 2e3 + 3e3)
        npt.assert_equal(self.outputs["platform_Awater"], 30)
        npt.assert_equal(self.outputs["platform_Iwaterx"], 6 * 15 + 5 * np.sum(R[:, 1] ** 2))
        npt.assert_equal(self.outputs["platform_Iwatery"], 6 * 15 + 5 * np.sum(R[:, 0] ** 2))
        npt.assert_equal(self.outputs["platform_added_mass"], 6 * np.arange(6))
        npt.assert_equal(self.outputs["platform_variable_capacity"], 10 + np.arange(6))
        npt.assert_equal(self.outputs["transition_piece_I"], 1e3 * 0.5**2 * np.r_[0.5, 0.5, 1.0, np.zeros(3)])
        npt.assert_array_less(1e2, self.outputs["platform_I_hull"] - self.outputs["transition_piece_I"])

    def testPlatformTurbineSystem(self):
        myobj = sys.PlatformFrame(options=self.opt)
        myobj.node_mem2glob = {}
        myobj.node_glob2mem = {}
        myobj.compute(self.inputs, self.outputs, self.discrete_inputs, self.discrete_outputs)
        for k in self.outputs:
            self.inputs[k] = self.outputs[k]
        for k in self.discrete_outputs:
            self.discrete_inputs[k] = self.discrete_outputs[k]
        myobj = sys.PlatformTurbineSystem()
        myobj.compute(self.inputs, self.outputs)

        npt.assert_equal(self.outputs["system_structural_mass"], 6e3 + 1e4 + 1e3)
        npt.assert_equal(
            self.outputs["system_structural_center_of_mass"],
            (6e3 * np.array([0.375, 0.25, 0.25]) + 1e3 * np.array([0.0, 0.0, 1.0]) + 1e4 * np.array([0.0, 0.0, 50.0]))
            / 1.7e4,
        )
        npt.assert_equal(self.outputs["variable_ballast_mass"], 6e4 - self.outputs["system_structural_mass"] - 3e3 / g)
        npt.assert_almost_equal(
            self.outputs["constr_variable_margin"],
            self.outputs["variable_ballast_mass"] / 1e3 / (10 + np.arange(6)).sum(),
        )
        frac = (10 + np.arange(6)) / (10 + np.arange(6)).sum()
        V_frac = self.outputs["variable_ballast_mass"] / 1e3 * frac
        npt.assert_almost_equal(self.outputs["member_variable_volume"], V_frac)
        s_cg = np.interp(0.5 * V_frac, np.arange(10), np.linspace(0, 0.5, 10))
        s_end = np.interp(V_frac, np.arange(10), np.linspace(0, 0.5, 10))
        npt.assert_almost_equal(self.outputs["member_variable_height"], s_end)
        cg_mem = np.zeros((6, 3))
        for k in range(6):
            cg_mem[k, :] = (
                s_cg[k] * np.diff(self.inputs[f"member{k}:nodes_xyz"][:2, :], axis=0)
                + self.inputs[f"member{k}:nodes_xyz"][0, :]
            )
        cg_var = np.dot(V_frac, cg_mem) / (self.outputs["variable_ballast_mass"] / 1e3)
        npt.assert_allclose(self.outputs["system_mass"], 6e3 + 1e4 + 1e3 + self.outputs["variable_ballast_mass"], 1e-6)
        npt.assert_almost_equal(
            self.outputs["system_center_of_mass"],
            (
                (6e3 + 1e4 + 1e3) * self.outputs["system_structural_center_of_mass"]
                + self.outputs["variable_ballast_mass"] * cg_var
            )
            / self.outputs["system_mass"],
        )
        npt.assert_array_less(self.outputs["platform_I_hull"], self.outputs["platform_I_total"])
        npt.assert_allclose(self.outputs["platform_mass"], 6e3 + 1e3 + self.outputs["variable_ballast_mass"], 1e-6)


class TestGroup(unittest.TestCase):
    def testRunAll(self):
        opt = {}
        opt["floating"] = {}
        opt["WISDEM"] = {}
        opt["WISDEM"]["FloatingSE"] = {}
        opt["floating"]["members"] = {}
        opt["floating"]["members"]["n_members"] = n_member = 6
        opt["floating"]["members"]["n_height"] = [2]
        opt["floating"]["members"]["n_bulkheads"] = [4]
        opt["floating"]["members"]["n_layers"] = [1]
        opt["floating"]["members"]["n_ballasts"] = [0]
        opt["floating"]["members"]["n_axial_joints"] = [1]
        opt["floating"]["members"]["outer_shape"] = n_member * ["circular"]
        opt["floating"]["rigid_bodies"] = {}
        opt["floating"]["rigid_bodies"]["n_bodies"] = 0
        opt["mooring"] = {}
        opt["mooring"]["n_attach"] = 3
        opt["mooring"]["n_anchors"] = 3

        opt["materials"] = {}
        opt["materials"]["n_mat"] = 2

        prob = om.Problem(reports=False)
        prob.model = sys.FloatingSystem(modeling_options=opt)
        prob.setup()

        prob["member0:nodes_xyz"][:2, :] = np.array([[0, 0, 0], [1, 0, 0]])
        prob["member1:nodes_xyz"][:2, :] = np.array([[1, 0, 0], [0.5, 1, 0]])
        prob["member2:nodes_xyz"][:2, :] = np.array([[0.5, 1, 0], [0, 0, 0]])
        prob["member3:nodes_xyz"][:2, :] = np.array([[0, 0, 0], [0, 0, 1]])
        prob["member4:nodes_xyz"][:2, :] = np.array([[1, 0, 0], [0, 0, 1]])
        prob["member5:nodes_xyz"][:2, :] = np.array([[0.5, 1, 0], [0, 0, 1]])
        for k in range(n_member):
            L = np.sqrt(np.sum(np.diff(prob[f"member{k}:nodes_xyz"][:, 2], axis=0) ** 2))
            prob[f"member{k}:nodes_r"][:2] = 0.1 * k * np.ones(2)
            prob[f"member{k}:section_D"][:1] = 2.0
            prob[f"member{k}:section_t"][:1] = 0.1
            prob[f"member{k}:section_A"][:1] = 0.5 * k * np.ones(1) + 1
            prob[f"member{k}:section_Asx"][:1] = 0.5 * k * np.ones(1) + 1
            prob[f"member{k}:section_Asy"][:1] = 0.5 * k * np.ones(1) + 1
            prob[f"member{k}:section_Ixx"][:1] = 2 * k * np.ones(1) + 1
            prob[f"member{k}:section_Iyy"][:1] = 2 * k * np.ones(1) + 1
            prob[f"member{k}:section_J0"][:1] = 2 * k * np.ones(1) + 1
            prob[f"member{k}:section_rho"][:1] = 1e3 / (0.5 * k * np.ones(1) + 1) / L
            prob[f"member{k}:section_E"][:1] = 3 * k * np.ones(1) + 1
            prob[f"member{k}:section_G"][:1] = 4 * k * np.ones(1) + 1
            prob[f"member{k}:section_TorsC"][:1] = 4 * k * np.ones(1) + 1
            prob[f"member{k}:section_sigma_y"][:1] = 5 * k * np.ones(1) + 1
            prob[f"member{k}:idx_cb"] = 0
            prob[f"member{k}:buoyancy_force"] = 1e2
            prob[f"member{k}:displacement"] = 1e1
            prob[f"member{k}:center_of_buoyancy"] = prob[f"member{k}:nodes_xyz"][:2, :].mean(axis=0)
            prob[f"member{k}:center_of_mass"] = prob[f"member{k}:nodes_xyz"][:2, :].mean(axis=0)
            prob[f"member{k}:total_mass"] = 1e3
            prob[f"member{k}:total_cost"] = 2e3
            prob[f"member{k}:Awater"] = 5.0
            prob[f"member{k}:Iwaterx"] = 15.0
            prob[f"member{k}:Iwatery"] = 15.0
            prob[f"member{k}:added_mass"] = np.arange(6)

        # Set environment to that used in OC3 testing campaign
        prob["rho_water"] = 1025.0  # Density of water [kg/m^3]

        prob["mooring_neutral_load"] = np.zeros((3, 3))
        prob["mooring_neutral_load"][:, 0] = [200, -100.0, -100]
        prob["mooring_neutral_load"][:, 1] = [0.0, 50, -50]
        prob["mooring_neutral_load"][:, 2] = -1e3
        prob["transition_piece_mass"] = 1e3
        prob["transition_piece_cost"] = 1e4
        prob["transition_node"] = prob["member4:nodes_xyz"][0, :]

        prob.run_model()
        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()



from wisdem.landbosse.model.CostModule import CostModule

import traceback
import numpy as np
import pandas as pd

class TurbineCost(CostModule):
    """
    **TurbineCost.py**

    Calculates the CapEx cost of the turbine itself

    This is a simple feedthrough calculation that requires an input CapEx assumption ($/kW)

    The structure of this class contains many superfluous elements, but it has been made to match
    the other LandBOSSE cost modules

    turbine_capex
        (float) CapEx cost of turbine ($/kW)

    number_turbines
        (int) Number of turbines
    """

    def __init__(self, input_dict, output_dict, project_name):
        """
        Parameters
        ----------
        input_dict : dict
            The input dictionary with key value pairs described in the
            class documentation

        output_dict : dict
            The output dictionary with key value pairs as found on the
            output documentation.
        """
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name
        
    def calculate_costs(self, calculate_costs_input_dict , calculate_costs_output_dict):
        """
        Function to calculate turbine cost in USD

        Parameters
        -------
        turbine_capex
            (in $/kW, default $1500/kW)

        number_turbines
            (unitless)

        Returns:
        -------
        turbine_cost
            (in USD)

        """

        turbine_capex = calculate_costs_input_dict["turbine_capex"]
        if np.isnan(turbine_capex):
            turbine_capex = 0.0

        num_turbines = calculate_costs_input_dict["num_turbines"]
        turbine_rating_kW = calculate_costs_input_dict["turbine_rating_MW"] * 1000

        turbine_cost = turbine_capex * num_turbines * turbine_rating_kW

        calculate_costs_output_dict["turbine_cost"] = turbine_cost

        calculate_costs_output_dict["turbine_cost_output_df"] = pd.DataFrame(
            [["Other", calculate_costs_output_dict["turbine_cost"], "Turbine"]],
            columns=["Type of cost", "Cost USD", "Phase of construction"],
        )


    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """

        result = []
        module = type(self).__name__

        result.append({
            "unit": "usd",
            "type": "variable",
            "variable_df_key_col_name": "Turbine CapEx",
            "value": float(self.output_dict["turbine_cost"]),
        })

        for _dict in result:
            _dict["project_id_with_serial"] = self.project_name
            _dict["module"] = module

        self.output_dict["turbine_cost_csv"] = result

    def run_module(self):
        """
        Runs the TurbineCost module and populates the IO dictionaries with calculated values.

         Parameters
        ----------
        <None>

        Returns
        -------
        tuple
            First element of tuple contains a 0 or 1. 0 means no errors happened and
            1 means an error happened and the module failed to run. The second element
            either returns a 0 if the module ran successfully, or it returns the error
            raised that caused the failure.

        """
        try:
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            self.output_dict["turbine_module_type_operation"] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict["turbine_cost_output_df"],
                project_id=self.project_name,
                total_or_turbine=True
            )
            return 0, 0
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} TurbineCost")
            return 1, error



"""
A test where we can modify different parts of the modeling_options to ensure
they work with a top-to-bottom WISDEM run.
"""

import os
import unittest

from wisdem.glue_code.runWISDEM import run_wisdem
import numpy.testing as npt

test_dir = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))),
    "examples", "11_user_custom")


class TestRegression(unittest.TestCase):
    def test15MW(self):
        ## IEA 15MW
        fname_wt_input = os.path.join(test_dir, "nrel5mw-spar_oc3_user_mass.yaml")
        fname_modeling_options = os.path.join(test_dir, "modeling_options.yaml")
        fname_analysis_options = os.path.join(test_dir, "analysis_options.yaml")
        wt_opt, modeling_options, opt_options = run_wisdem(
            fname_wt_input, fname_modeling_options, fname_analysis_options
        )

        self.assertEqual(wt_opt["drivese.hub_mass"][0], 2700.0)
        self.assertEqual(wt_opt["drivese.spinner_mass"][0], 500.0)
        self.assertEqual(wt_opt["drivese.pitch_mass"][0], 8300.0)
        self.assertEqual(wt_opt["drivese.mb1_mass"][0], 1500.0)
        self.assertEqual(wt_opt["drivese.mb2_mass"][0], 1350.0)
        self.assertEqual(wt_opt["drivese.bedplate_mass"][0], 20000.0)
        self.assertEqual(wt_opt["drivese.brake_mass"][0], 5500.0)
        self.assertEqual(wt_opt["drivese.converter_mass"][0], 4200.0)
        self.assertEqual(wt_opt["drivese.transformer_mass"][0], 11500.0)
        self.assertEqual(wt_opt["drivese.gearbox_mass"][0], 21500.0)
        self.assertEqual(wt_opt["drivese.generator_mass"][0], 19500.0)
        self.assertEqual(wt_opt["towerse.tower_mass"][0], 250000.0)
        self.assertAlmostEqual(wt_opt["floatingse.platform_hull_mass"][0], 203000.0)


    def testElastic(self):
        fname_wt_input = os.path.join(test_dir, "IEA-15-240-RWT_VolturnUS-S_user_elastic.yaml")
        fname_modeling_options = os.path.join(test_dir, "modeling_options_user_elastic.yaml")
        fname_analysis_options = os.path.join(test_dir, "analysis_options_null.yaml")
        wt_opt, modeling_options, opt_options = run_wisdem(
            fname_wt_input, fname_modeling_options, fname_analysis_options
        )

        npt.assert_almost_equal(wt_opt['drivese.hub_system_mass'][0], 69360.0)
        npt.assert_almost_equal(wt_opt['drivese.hub_system_I'], [973520., 619970., 619970., 0.0, 0.0, 0.0])
        npt.assert_almost_equal(wt_opt['drivese.generator_rotor_I'], [9912933., 10862815., 10360761.])
        npt.assert_almost_equal(wt_opt['drivese.above_yaw_mass'][0], 675175.0)
        npt.assert_almost_equal(wt_opt['drivese.yaw_mass'][0], 0)
        npt.assert_almost_equal(wt_opt['drivese.above_yaw_cm'], [-4.528, -0.14 ,  4.098])
        npt.assert_almost_equal(wt_opt['drivese.above_yaw_I_TT'], [ 9912933., 10862815., 10360761., 0.0, 0.0, 0.0])
        npt.assert_almost_equal(wt_opt['drivese.above_yaw_I'], [ 9912933., 10862815., 10360761., 0.0, 0.0, 0.0])
        npt.assert_almost_equal(wt_opt['nacelle.uptilt'][0], 0.10471976)
        npt.assert_almost_equal(wt_opt['nacelle.overhang'][0], 12.032)
        npt.assert_almost_equal(wt_opt['nacelle.distance_tt_hub'][0], 5.6141)
        
if __name__ == "__main__":
    unittest.main()



from wisdem.landbosse.model.ManagementCost import ManagementCost
from wisdem.landbosse.model.Manager import Manager
from wisdem.landbosse.model.WeatherDelay import WeatherDelay
from wisdem.landbosse.model.FoundationCost import FoundationCost
from wisdem.landbosse.model.ErectionCost import ErectionCost
from wisdem.landbosse.model.SitePreparationCost import SitePreparationCost
from wisdem.landbosse.model.SubstationCost import SubstationCost
from wisdem.landbosse.model.GridConnectionCost import GridConnectionCost
from wisdem.landbosse.model.CollectionCost import Cable, Array, ArraySystem
from wisdem.landbosse.model.DevelopmentCost import DevelopmentCost
from wisdem.landbosse.model.DefaultMasterInputDict import DefaultMasterInputDict



import os
import numpy as np
import multiprocessing as mp
import time
from wisdem import run_wisdem

parallel_flag = True

def parallel_runner(k, fgeom, fmodel, fanal, dover):
    iwt_opt, _, _ = run_wisdem(fgeom, fmodel, fanal, overridden_values=dover)
    return k, float(iwt_opt['financese.turbine_aep'][0])

def driver():
    start = time.time()

    # File management
    mydir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))  # get path to examples dir
    fname_wt_input = os.path.join(mydir, "02_reference_turbines", "IEA-15-240-RWT.yaml")
    fname_modeling_options = os.path.join(mydir, "02_reference_turbines", "modeling_options.yaml")
    fname_analysis_options = os.path.join(mydir, "02_reference_turbines", "analysis_options.yaml")

    # Initial run
    wt_opt, modeling_options, analysis_options = run_wisdem(fname_wt_input, fname_modeling_options,
                                                            fname_analysis_options)

    # Set parametric values
    blade_cones = np.deg2rad( np.arange(0, 5, 2) )
    shaft_tilts = np.deg2rad( np.arange(0, 5, 2) )

    # Set to grid of points and then sequence of run values
    Blades, Shafts = np.meshgrid(blade_cones, shaft_tilts)
    Blades = Blades.flatten()
    Shafts = Shafts.flatten()
    npts = Blades.size

    # Run parametric loop with overrides
    aep_output = np.zeros( npts ) # Initialize output container
    myargs = [] # Container for parallel run arguments
    for k in range(npts):
        overrides = {'hub.cone':Blades[k], 'nacelle.uptilt': Shafts[k]}
        myargs.append([k, fname_wt_input, fname_modeling_options, fname_analysis_options, overrides])

        # Run cases in serial this way
        if not parallel_flag:
            iwt_opt, _, _ = run_wisdem(fname_wt_input, fname_modeling_options,
                                       fname_analysis_options, overridden_values=overrides)
            aep_output[k] = float(iwt_opt['financese.turbine_aep'][0])

    # Run cases in parallel this way
    if parallel_flag:
        ncore = max(1, mp.cpu_count() - 2)
        pool = mp.Pool(processes=ncore)
        results = pool.starmap(parallel_runner, myargs)
        for k in results:
            aep_output[ k[0] ] = k[1]
    
    print(aep_output)
    finish = time.time()
    print((finish-start)/60.0)

if __name__ == "__main__":
    driver()
    



# TODO: Code commenting and RST parallel

import numpy as np
import openmdao.api as om

from wisdem.commonse import fileIO
from wisdem.floatingse import FloatingSE

plot_flag = False  # True
opt_flag = False

npts = 5

opt = {}
opt["flags"] = {}
opt["flags"]["floating"] = opt["flags"]["offshore"] = opt["flags"]["tower"] = True
opt["flags"]["tower"] = False
opt["floating"] = {}
opt["floating"]["members"] = {}
opt["floating"]["members"]["n_members"] = 4
opt["floating"]["members"]["n_height"] = [npts, 3, 3, 3]
opt["floating"]["members"]["n_bulkheads"] = [4, 2, 2, 2]
opt["floating"]["members"]["n_layers"] = [1, 1, 1, 1]
opt["floating"]["members"]["n_ballasts"] = [2, 0, 0, 0]
opt["floating"]["members"]["n_axial_joints"] = [0, 0, 0, 0]
opt["floating"]["members"]["outer_shape"] = 4 * ["circular"]
opt["floating"]["rigid_bodies"] = {}
opt["floating"]["rigid_bodies"]["n_bodies"] = 0
opt["WISDEM"] = {}
opt["WISDEM"]["n_dlc"] = 1
opt["WISDEM"]["TowerSE"] = {}
opt["WISDEM"]["TowerSE"]["n_height"] = npts
opt["WISDEM"]["TowerSE"]["n_layers"] = 1
opt["WISDEM"]["TowerSE"]["n_refine"] = 1
opt["WISDEM"]["FloatingSE"] = {}
opt["WISDEM"]["FloatingSE"]["frame3dd"] = {}
opt["WISDEM"]["FloatingSE"]["frame3dd"]["shear"] = False
opt["WISDEM"]["FloatingSE"]["frame3dd"]["geom"] = False
opt["WISDEM"]["FloatingSE"]["frame3dd"]["modal_method"] = 2
opt["WISDEM"]["FloatingSE"]["frame3dd"]["tol"] = 1e-6
opt["WISDEM"]["FloatingSE"]["gamma_f"] = 1.35  # Safety factor on loads
opt["WISDEM"]["FloatingSE"]["gamma_m"] = 1.3  # Safety factor on materials
opt["WISDEM"]["FloatingSE"]["gamma_n"] = 1.0  # Safety factor on consequence of failure
opt["WISDEM"]["FloatingSE"]["gamma_b"] = 1.1  # Safety factor on buckling
opt["WISDEM"]["FloatingSE"]["gamma_fatigue"] = 1.755  # Not used
opt["WISDEM"]["FloatingSE"]["rank_and_file"] = True
opt["mooring"] = {}
opt["mooring"]["n_attach"] = 3
opt["mooring"]["n_anchors"] = 3
opt["mooring"]["line_anchor"] = ["custom"] * 3
opt["mooring"]["line_material"] = ["custom"] * 3
opt["materials"] = {}
opt["materials"]["n_mat"] = 2

prob = om.Problem(reports=False)
prob.model = FloatingSE(modeling_options=opt)
prob.setup()

# Material properties
prob["rho_mat"] = np.array([7850.0, 5000.0])  # Steel, ballast slurry [kg/m^3]
prob["E_mat"] = 200e9 * np.ones((2, 3))  # Young's modulus [N/m^2]
prob["G_mat"] = 79.3e9 * np.ones((2, 3))  # Shear modulus [N/m^2]
prob["sigma_y_mat"] = 3.45e8 * np.ones(2)  # Elastic yield stress [N/m^2]
prob["sigma_ult_mat"] = 5e8 * np.ones((2, 3))
prob["wohler_exp_mat"] = 4.0 * np.ones(2)
prob["wohler_A_mat"] = 7.5e8 * np.ones(2)
prob["unit_cost_mat"] = np.array([2.0, 1.0])
prob["material_names"] = ["steel", "slurry"]

# Mass and cost scaling factors
prob["labor_cost_rate"] = 1.0  # Cost factor for labor time [$/min]
prob["painting_cost_rate"] = 14.4  # Cost factor for column surface finishing [$/m^2]

# Main geometry
h = np.array([10.0, 20.0, 10.0, 8.0])
prob["member0.outfitting_factor_in"] = 1.05  # Fraction of additional outfitting mass for each column
prob["member0.grid_axial_joints"] = []
prob["member0.ballast_grid"] = np.array([[0, 0.25], [0, 0.5]])
prob["member0.ballast_volume"] = [np.pi * 7**2 * 5, 0.0]
prob["member0.s_in"] = np.cumsum(np.r_[0, h]) / h.sum()
prob["member0.outer_diameter_in"] = 14 * np.ones(npts)
prob["member0.layer_thickness"] = 0.05 * np.ones((1, npts))
prob["member0.layer_materials"] = ["steel"]
prob["member0.ballast_materials"] = ["slurry", "seawater"]
prob["member0:joint1"] = np.array([0.0, 0.0, 8.0 - h.sum()])
prob["member0:joint2"] = np.array([0.0, 0.0, 8.0])  # Freeboard=10
prob["member0.bulkhead_thickness"] = 0.05 * np.ones(4)  # Locations of internal bulkheads
prob["member0.bulkhead_grid"] = np.array([0.0, 0.25, 0.5, 1.0])
prob["member0.ring_stiffener_web_height"] = 0.10
prob["member0.ring_stiffener_web_thickness"] = 0.04
prob["member0.ring_stiffener_flange_width"] = 0.10
prob["member0.ring_stiffener_flange_thickness"] = 0.02
prob["member0.ring_stiffener_spacing"] = 0.044791667  # non-dimensional ring stiffener spacing

# Now do the legs
angs = np.linspace(0, 2 * np.pi, 1 + opt["mooring"]["n_attach"])
for k in range(1, 4):
    prob["member" + str(k) + ".outfitting_factor_in"] = 1.05  # Fraction of additional outfitting mass for each column
    prob["member" + str(k) + ".grid_axial_joints"] = []
    prob["member" + str(k) + ".s_in"] = np.array([0.0, 0.5, 1.0])
    prob["member" + str(k) + ".outer_diameter_in"] = 5 * np.ones(3)
    prob["member" + str(k) + ".ca_usr_grid"] = 2.0*np.ones(3)  # Added mass coefficient
    prob["member" + str(k) + ".cd_usr_grid"] = -1.0*np.ones(3)  # drag coefficient
    prob["member" + str(k) + ".layer_thickness"] = 0.05 * np.ones((1, 3))
    prob["member" + str(k) + ".layer_materials"] = ["steel"]
    prob["member" + str(k) + ".ballast_materials"] = []
    prob["member" + str(k) + ":joint1"] = np.array([30.0 * np.cos(angs[k - 1]), 30.0 * np.sin(angs[k - 1]), -40.0])
    prob["member" + str(k) + ":joint2"] = np.array([0.0, 0.0, -40.0])  # Freeboard=10
    prob["member" + str(k) + ".bulkhead_thickness"] = 0.05 * np.ones(2)  # Locations of internal bulkheads
    prob["member" + str(k) + ".bulkhead_grid"] = np.array([0.0, 1.0])
    prob["member" + str(k) + ".ring_stiffener_web_height"] = 0.10
    prob["member" + str(k) + ".ring_stiffener_web_thickness"] = 0.04
    prob["member" + str(k) + ".ring_stiffener_flange_width"] = 0.10
    prob["member" + str(k) + ".ring_stiffener_flange_thickness"] = 0.02
    prob["member" + str(k) + ".ring_stiffener_spacing"] = 0.06666667  # non-dimensional ring stiffener spacing

# Mooring parameters: Nylon
prob["line_diameter"] = 0.5  # Diameter of mooring line/chain [m]
prob["line_length"] = 250.0  # Unstretched mooring line length
prob["line_mass_density_coeff"] = 0.6476e3
prob["line_stiffness_coeff"] = 1.18e8
prob["line_breaking_load_coeff"] = 139357e3
prob["line_cost_rate_coeff"] = 3.415e4
prob["fairlead_radius"] = 30.0
prob["fairlead"] = 40.0
prob["anchor_radius"] = 50.0
prob["anchor_cost"] = 1e5

# Mooring constraints
prob["max_surge_fraction"] = 0.1  # Max surge/sway offset [m]
prob["survival_heel"] = 10.0  # Max heel (pitching) angle [deg]
prob["operational_heel"] = 5.0  # Max heel (pitching) angle [deg]

# Set environment to that used in OC3 testing campaign
prob["rho_air"] = 1.226  # Density of air [kg/m^3]
prob["mu_air"] = 1.78e-5  # Viscosity of air [kg/m/s]
prob["rho_water"] = 1025.0  # Density of water [kg/m^3]
prob["mu_water"] = 1.08e-3  # Viscosity of water [kg/m/s]
prob["water_depth"] = 320.0  # Distance to sea floor [m]
prob["Hsig_wave"] = 10.8  # Significant wave height [m]
prob["Tsig_wave"] = 9.8  # Wave period [s]
prob["shearExp"] = 0.11  # Shear exponent in wind power law
prob["Uc"] = 0.0  # Mean current speed
prob["beta_wind"] = prob["beta_wave"] = 0.0
prob["env.Uref"] = 11.0
prob["wind_reference_height"] = 119.0

prob["transition_node"] = prob["member0:joint2"]

# Properties of rotor-nacelle-assembly (RNA)
prob["turbine_mass"] = 350e3
prob["turbine_F"] = np.array([1284744.196, 0, -112400.5527])
prob["turbine_M"] = np.array([3963732.762, 896380.8464, -346781.682])

# Use FD and run optimization
prob.run_model()
prob.model.list_outputs(units=True)

# Visualize with mayavi, which can be difficult to install
if plot_flag:
    import wisdem.floatingse.visualize as viz

    vizobj = viz.Visualize(prob)
    vizobj.draw_spar()
